from adsg_core.optimization.assign_enc.patterns import *
from adsg_core.optimization.assign_enc.enumerating import *
from adsg_core.optimization.assign_enc.lazy.encodings import *
from adsg_core.optimization.assign_enc.lazy.imputation import *
from adsg_core.optimization.assign_enc.eager.encodings import *
from adsg_core.optimization.assign_enc.eager.imputation import *


EAGER_ENCODERS = [
    lambda imp: DirectMatrixEncoder(imp),
    # lambda imp: DirectMatrixEncoder(imp, remove_gaps=False),
    lambda imp: ElementGroupedEncoder(imp),
    # lambda imp: ElementGroupedEncoder(imp, normalize_within_group=False),
    lambda imp: ConnIdxGroupedEncoder(imp),
    lambda imp: ConnIdxGroupedEncoder(imp, by_src=False),
    lambda imp: ConnIdxGroupedEncoder(imp, binary=True),
    lambda imp: ConnIdxGroupedEncoder(imp, by_src=False, binary=True),

    lambda imp: AmountFirstGroupedEncoder(imp, TotalAmountGrouper(), OneVarLocationGrouper()),
    lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountGrouper(), OneVarLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountFlattenedGrouper(), OneVarLocationGrouper()),
    lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountGrouper(), OneVarLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountFlattenedGrouper(), OneVarLocationGrouper()),

    # lambda imp: AmountFirstGroupedEncoder(imp, TotalAmountGrouper(), FlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountGrouper(), FlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountFlattenedGrouper(), FlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountGrouper(), FlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountFlattenedGrouper(), FlatIndexLocationGrouper()),

    # lambda imp: AmountFirstGroupedEncoder(imp, TotalAmountGrouper(), RelFlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountGrouper(), RelFlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountFlattenedGrouper(), RelFlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountGrouper(), RelFlatIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountFlattenedGrouper(), RelFlatIndexLocationGrouper()),

    # lambda imp: AmountFirstGroupedEncoder(imp, TotalAmountGrouper(), CoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountGrouper(), CoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountFlattenedGrouper(), CoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountGrouper(), CoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountFlattenedGrouper(), CoordIndexLocationGrouper()),

    # lambda imp: AmountFirstGroupedEncoder(imp, TotalAmountGrouper(), RelCoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountGrouper(), RelCoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, SourceAmountFlattenedGrouper(), RelCoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountGrouper(), RelCoordIndexLocationGrouper()),
    # lambda imp: AmountFirstGroupedEncoder(imp, TargetAmountFlattenedGrouper(), RelCoordIndexLocationGrouper()),
]

DEFAULT_EAGER_ENCODER = lambda: DirectMatrixEncoder(DEFAULT_EAGER_IMPUTER())

EAGER_IMPUTERS = [
    lambda: FirstImputer(),
    lambda: AutoModImputer(),
    lambda: AutoModImputer(reverse=True),
    lambda: DeltaImputer(),
    lambda: ClosestImputer(),
    lambda: ClosestImputer(euclidean=False),
    lambda: ConstraintViolationImputer(),
]

DEFAULT_EAGER_IMPUTER = AutoModImputer

LAZY_ENCODERS = [
    lambda imp: LazyDirectMatrixEncoder(imp),

    lambda imp: LazyAmountFirstEncoder(imp, FlatLazyAmountEncoder(), FlatLazyConnectionEncoder()),
    # lambda imp: LazyAmountFirstEncoder(imp, TotalLazyAmountEncoder(), FlatLazyConnectionEncoder()),
    # lambda imp: LazyAmountFirstEncoder(imp, SourceLazyAmountEncoder(), FlatLazyConnectionEncoder()),
    lambda imp: LazyAmountFirstEncoder(imp, SourceTargetLazyAmountEncoder(), FlatLazyConnectionEncoder()),

    lambda imp: LazyConnIdxMatrixEncoder(imp, FlatConnCombsEncoder()),
    lambda imp: LazyConnIdxMatrixEncoder(imp, FlatConnCombsEncoder(), by_src=False),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, FlatConnCombsEncoder(), amount_first=True),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, FlatConnCombsEncoder(), by_src=False, amount_first=True),
    lambda imp: LazyConnIdxMatrixEncoder(imp, GroupedConnCombsEncoder()),
    lambda imp: LazyConnIdxMatrixEncoder(imp, GroupedConnCombsEncoder(), by_src=False),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, GroupedConnCombsEncoder(), amount_first=True),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, GroupedConnCombsEncoder(), by_src=False, amount_first=True),
    lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder()),
    lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(), by_src=False),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(), amount_first=True),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(), by_src=False, amount_first=True),
    lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(binary=True)),
    lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(binary=True), by_src=False),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(binary=True), amount_first=True),
    # lambda imp: LazyConnIdxMatrixEncoder(imp, ConnIdxCombsEncoder(binary=True), by_src=False, amount_first=True),
]

DEFAULT_LAZY_ENCODER = lambda: LazyDirectMatrixEncoder(DEFAULT_LAZY_IMPUTER())

EAGER_ENUM_ENCODERS = [  # Encoders only based on counting the possibilities, ignoring the actual connection patterns
    lambda imp: EnumOrdinalEncoder(imp),
    lambda imp: EnumRecursiveEncoder(imp, n_divide=2),
    lambda imp: EnumRecursiveEncoder(imp, n_divide=3),
    lambda imp: EnumRecursiveEncoder(imp, n_divide=4),
]

PATTERN_ENCODERS = [  # Encoders for specific assignment patterns
    lambda imp: CombiningPatternEncoder(imp),
    lambda imp: AssigningPatternEncoder(imp),
    lambda imp: PartitioningPatternEncoder(imp),
    lambda imp: ConnectingPatternEncoder(imp),
    lambda imp: PermutingPatternEncoder(imp),
    lambda imp: UnorderedCombiningPatternEncoder(imp),
]

LAZY_IMPUTERS = [
    lambda: LazyFirstImputer(),
    lambda: LazyDeltaImputer(),
    lambda: LazyClosestImputer(),
    lambda: LazyClosestImputer(euclidean=False),
    lambda: LazyConstraintViolationImputer(),
]

DEFAULT_LAZY_IMPUTER = LazyDeltaImputer



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.matrix import *
from adsg_core.optimization.assign_enc.encoding import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['AssignmentManagerBase', 'AssignmentManager', 'LazyAssignmentManager']

T = TypeVar('T')


class AssignmentManagerBase:
    """Base class for managing the encoding of assignment problems."""

    def get_random_design_vector(self) -> DesignVector:
        return [dv.get_random() for dv in self.design_vars]

    @staticmethod
    def _is_violated_matrix(matrix: np.ndarray) -> bool:
        """Whether the matrix represents output of the constraint violator imputer"""
        try:
            return matrix[0, 0] == X_INACTIVE_VALUE
        except IndexError:
            return False

    @staticmethod
    def _correct_is_active(vector: DesignVector) -> Tuple[DesignVector, IsActiveVector]:
        """Corrects the design vector (replace -1 with 0) and return is_active vector"""
        corrected_vector = np.array(vector)
        is_active = corrected_vector != X_INACTIVE_VALUE
        corrected_vector[corrected_vector == X_INACTIVE_VALUE] = 0
        return corrected_vector, is_active

    @property
    def encoder(self) -> Encoder:
        raise NotImplementedError

    @property
    def matrix_gen(self) -> AggregateAssignmentMatrixGenerator:
        raise NotImplementedError

    @property
    def design_vars(self):
        raise NotImplementedError

    def get_imputation_ratio(self) -> float:
        """Ratio of the total design space size to the actual amount of possible connections"""
        raise NotImplementedError

    def correct_vector(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector]:
        """Correct the design vector so that it matches an existing connection pattern"""
        raise NotImplementedError

    def get_matrix(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, np.ndarray]:
        """Get connection matrix from a given design vector"""
        raise NotImplementedError

    def get_conn_idx(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, Optional[List[Tuple[int, int]]]]:
        """Get node connections for a given design vector"""
        raise NotImplementedError

    def get_conns(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, Optional[List[Tuple[Node, Node]]]]:
        """Get node connections for a given design vector"""
        raise NotImplementedError

    def get_all_design_vectors(self) -> Dict[NodeExistence, np.ndarray]:
        """Returns all possible valid design vectors, with inactive design variables having a value of -1"""
        raise NotImplementedError


class AssignmentManager(AssignmentManagerBase):
    """
    Uses the matrix generator and encoder to implement the conversion between design vector and connection pattern.
    """

    def __init__(self, settings: MatrixGenSettings, encoder: EagerEncoder, cache=True):
        self._matrix_gen = gen = AggregateAssignmentMatrixGenerator(settings)
        self._encoder = encoder

        encoder.matrix = gen.get_agg_matrix(cache=cache)

    @property
    def matrix_gen(self) -> AggregateAssignmentMatrixGenerator:
        return self._matrix_gen

    @property
    def encoder(self) -> EagerEncoder:
        return self._encoder

    @property
    def matrix(self) -> Dict[NodeExistence, np.ndarray]:
        return self._encoder.matrix

    @property
    def design_vars(self):
        return self._encoder.design_vars

    def get_imputation_ratio(self) -> float:
        return self._encoder.get_imputation_ratio()

    def correct_vector(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector]:
        """Correct the design vector so that it matches an existing connection pattern"""
        imputed_vector, _ = self._encoder.get_matrix(vector, existence=existence)
        return self._correct_is_active(imputed_vector)

    def get_matrix(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, np.ndarray]:
        """Get connection matrix from a given design vector"""
        imputed_vector, matrix = self._encoder.get_matrix(vector, existence=existence)
        imputed_vector, is_active = self._correct_is_active(imputed_vector)
        return imputed_vector, is_active, matrix

    def get_conn_idx(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, Optional[List[Tuple[int, int]]]]:
        """Get node connections for a given design vector"""

        # Get matrix and impute design vector
        imputed_vector, is_active, matrix = self.get_matrix(vector, existence=existence)
        if self._is_violated_matrix(matrix):
            return imputed_vector, is_active, None

        # Get connections from matrix
        edges_idx = self._matrix_gen.get_conn_idx(matrix)
        return imputed_vector, is_active, edges_idx

    def get_conns(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, Optional[List[Tuple[Node, Node]]]]:
        """Get node connections for a given design vector"""

        # Get matrix and impute design vector
        imputed_vector, is_active, matrix = self.get_matrix(vector, existence=existence)
        if self._is_violated_matrix(matrix):
            return imputed_vector, is_active, None

        # Get connections from matrix
        edges = self._matrix_gen.get_conns(matrix)
        return imputed_vector, is_active, edges

    def get_all_design_vectors(self) -> Dict[NodeExistence, np.ndarray]:
        return self.encoder.padded_design_vectors


class LazyAssignmentManager(AssignmentManagerBase):

    def __init__(self, settings: MatrixGenSettings, encoder: LazyEncoder):
        self._encoder = encoder
        encoder.set_settings(settings)

    @property
    def encoder(self):
        return self._encoder

    @property
    def matrix_gen(self) -> AggregateAssignmentMatrixGenerator:
        return self._encoder.matrix_gen

    @property
    def design_vars(self):
        return self._encoder.design_vars

    def get_imputation_ratio(self) -> float:
        return self._encoder.get_imputation_ratio()

    def correct_vector(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector]:
        """Correct the design vector so that it matches an existing connection pattern"""
        imputed_vector, _ = self._encoder.get_matrix(vector, existence=existence)
        return self._correct_is_active(imputed_vector)

    def get_matrix(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, np.ndarray]:
        """Get connection matrix from a given design vector"""
        imputed_vector, matrix = self._encoder.get_matrix(vector, existence=existence)
        imputed_vector, is_active = self._correct_is_active(imputed_vector)
        return imputed_vector, is_active, matrix

    def get_conn_idx(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, Optional[List[Tuple[int, int]]]]:
        """Get node connections for a given design vector"""
        imputed_vector, matrix = self._encoder.get_matrix(vector, existence=existence)
        imputed_vector, is_active = self._correct_is_active(imputed_vector)
        if self._is_violated_matrix(matrix):
            return imputed_vector, is_active, None

        # Get connections from matrix
        edges_idx = self._encoder.get_conn_idx(matrix)
        return imputed_vector, is_active, edges_idx

    def get_conns(self, vector: DesignVector, existence: NodeExistence = None) \
            -> Tuple[DesignVector, IsActiveVector, Optional[List[Tuple[Node, Node]]]]:
        """Get node connections for a given design vector"""
        imputed_vector, matrix = self._encoder.get_matrix(vector, existence=existence)
        imputed_vector, is_active = self._correct_is_active(imputed_vector)
        if self._is_violated_matrix(matrix):
            return imputed_vector, is_active, None

        # Get connections from matrix
        edges = self._encoder.get_conns(matrix)
        return imputed_vector, is_active, edges

    def get_all_design_vectors(self) -> Dict[NodeExistence, np.ndarray]:
        return self._encoder.get_all_design_vectors()



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import itertools
from typing import *
import networkx as nx
from adsg_core.graph.traversal import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *

__all__ = ['add_incompatibility_constraint', 'get_confirmed_incompatibility_edges', 'IncompatibilityError',
           'get_mod_nodes_remove_incompatibilities', 'get_incompatibility_deriving_nodes']


def add_incompatibility_constraint(graph: nx.MultiDiGraph, nodes: List[DSGNode], **attr):
    # Loop over all pairs (both directions)
    for source_node, target_node in itertools.permutations(nodes, 2):
        add_edge(graph, source_node, target_node, edge_type=EdgeType.INCOMPATIBILITY, **attr)


def _get_canonical_edge(edge: EdgeTuple) -> EdgeTuple:
    source_node, target_node = sorted(edge[:2], key=lambda n: getattr(n, 'name', str(hash(n))))
    return source_node, target_node, 0, edge[-1]


def get_confirmed_incompatibility_edges(graph: nx.MultiDiGraph, start_nodes: Set[DSGNode]) -> Set[EdgeTuple]:
    """Confirmed incompatibility edges represent incompatibility constraints between confirmed nodes, and thereby
    show that the confirmed architecture is infeasible (as there is an incompatibility)."""
    confirmed_nodes, _ = traverse_until_choice_nodes(graph, start_nodes)

    edges: Set[EdgeTuple] = set()
    for edge in iter_edges(graph):
        if get_edge_type(edge) != EdgeType.INCOMPATIBILITY:
            continue
        if edge[0] in confirmed_nodes or edge[1] in confirmed_nodes:
            edges.add(_get_canonical_edge(edge))
    return edges


class IncompatibilityError(RuntimeError):

    def __init__(self, msg: str, edges: Set[EdgeTuple], removed_nodes: Set[DSGNode]):
        super(IncompatibilityError, self).__init__(msg)
        self.edges = {_get_canonical_edge(edge) for edge in edges}
        self.removed_nodes = removed_nodes

    def print_edges(self):
        return ', '.join([f'{edge[0]!s} <-> {edge[1]!s}' for edge in self.edges])


def get_mod_nodes_remove_incompatibilities(
        graph: nx.MultiDiGraph, start_nodes: Set[DSGNode], removed_edges: Set[EdgeTuple] = None, cache=None) \
        -> Set[DSGNode]:
    confirmed_nodes, _ = traverse_until_choice_nodes(graph, start_nodes)

    removed_nodes: Set[DSGNode] = set()
    confirmed_incompatibility_edges: Set[EdgeTuple] = set()
    infeasible_incompatibility_edges: Set[EdgeTuple] = set()
    for edge in iter_edges(graph):
        if get_edge_type(edge) != EdgeType.INCOMPATIBILITY:
            continue

        # If both nodes are in the confirmed nodes, the graph is infeasible
        if edge[0] in confirmed_nodes and edge[1] in confirmed_nodes:
            infeasible_incompatibility_edges.add(edge)
            continue

        # If the source node is confirmed, we need to remove target (and derived) incompatible nodes
        if edge[0] in confirmed_nodes:
            confirmed_incompatibility_edges.add(edge)
            removed_nodes.add(edge[1])

    if len(infeasible_incompatibility_edges) > 0:
        raise IncompatibilityError('Could not resolve incompatibility constraints', infeasible_incompatibility_edges,
                                   removed_nodes)

    if len(confirmed_incompatibility_edges) == 0:
        return set()

    # Remove nodes derived from incompatible target nodes
    if removed_edges is None:
        removed_edges = set()
    for edge in confirmed_incompatibility_edges:
        # If this is the target edge, do not remove the derived nodes
        if edge[1] in confirmed_nodes:
            continue

        # Additionally remove nodes that are derived by the incompatible node
        derived_edges, derived_nodes = \
            get_derived_edges_for_edge(graph, edge, start_nodes, removed_edges=removed_edges,
                                       removed_nodes=removed_nodes, cache=cache)
        removed_edges |= derived_edges
        removed_nodes |= derived_nodes

        # Additionally remove nodes that are deriving the incompatible node
        deriving_nodes = get_incompatibility_deriving_nodes(graph, edge[1], confirmed_nodes, cache=cache)
        removed_nodes |= deriving_nodes

        # And the nodes that are derived by the deriving nodes
        for deriving_node in deriving_nodes:
            deriving_derived_edges, deriving_derived_nodes =\
                get_derived_edges_for_node(graph, deriving_node, start_nodes, removed_edges=removed_edges,
                                           removed_nodes=removed_nodes, traversed=removed_nodes, cache=cache)
            removed_edges |= deriving_derived_edges
            removed_nodes |= deriving_derived_nodes

        # If any of the deriving nodes are in the confirmed nodes, we have an infeasible graph
        if len(deriving_nodes & confirmed_nodes) > 0:
            removed_nodes -= confirmed_nodes
            raise IncompatibilityError('Incompatibility constraint derives from confirmed nodes', {edge}, removed_nodes)

    return removed_nodes


def get_incompatibility_deriving_nodes(
        graph: nx.MultiDiGraph, target_node: DSGNode, confirmed_nodes: Set[DSGNode],
        _deriving_nodes: Set[DSGNode] = None, cache=None) -> Set[DSGNode]:

    deriving_nodes = {target_node}
    if _deriving_nodes is not None:
        deriving_nodes |= _deriving_nodes

    # Get deriving nodes
    option_decision_nodes: Set[SelectionChoiceNode] = set()
    for edge in iter_in_edges_cached(graph, target_node, cache=cache):
        if get_edge_type(edge) != EdgeType.DERIVES:
            continue
        deriving_node = edge[0]
        if deriving_node in deriving_nodes:
            continue

        # If it is a confirmed node, do not continue searching (as the calling function will fail anyway)
        if deriving_node in confirmed_nodes:
            deriving_nodes.add(deriving_node)
            continue

        # If it is a decision node, do not continue searching for additional deriving nodes
        if isinstance(deriving_node, SelectionChoiceNode):
            option_decision_nodes.add(deriving_node)
            continue

        deriving_nodes.add(deriving_node)
        deriving_nodes |= get_incompatibility_deriving_nodes(
            graph, deriving_node, confirmed_nodes, _deriving_nodes=deriving_nodes, cache=cache)

    # Check option-decision nodes
    for option_decision_node in option_decision_nodes:
        option_nodes = {edge[1] for edge in iter_out_edges_cached(graph, option_decision_node, cache=cache)
                        if get_edge_type(edge) == EdgeType.DERIVES}

        # If all option nodes are deriving nodes, it means that the option-decision would be left without any options,
        # and therefore we also need to remove this decision and its deriving nodes
        if len(option_nodes.difference(deriving_nodes)) == 0:
            deriving_nodes |= get_incompatibility_deriving_nodes(graph, option_decision_node, confirmed_nodes,
                                                                 _deriving_nodes=deriving_nodes, cache=cache)

    return deriving_nodes



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import enum
import random
from typing import *
from adsg_core.graph.adsg_nodes import MetricNode, DesignVariableNode, ChoiceNode

__all__ = ['DesVar', 'Direction', 'Objective', 'Constraint']


class DesVar:
    """
    Class representing a design variable. A design variable can either be discrete (options are specified) or continuous
    (bounds are specified).
    """

    def __init__(self, name: str, options: list = None, bounds: Tuple[float, float] = None,
                 node: Union[DesignVariableNode, ChoiceNode] = None, conditionally_active=False):
        if (options is None) == (bounds is None):
            raise ValueError('Either options or bounds must be provided: %s' % name)
        if options is not None:
            if len(options) == 0:
                raise ValueError('At least one option should be provided: %s' % name)
        if bounds is not None:
            if len(bounds) != 2:
                raise ValueError('Bounds should be a tuple: %s' % name)
            if bounds[0] >= bounds[1]:
                raise ValueError('Lower bound should be lower than upper bound: %.2f < %.2f (%s)' %
                                 (bounds[0], bounds[1], name))

        self._name = name
        self._opts = options
        self._bounds = bounds
        self._node = node
        self.conditionally_active = conditionally_active

    @classmethod
    def from_des_var_node(cls, des_var_node: DesignVariableNode, conditionally_active=False) -> 'DesVar':
        name = des_var_node.name
        if des_var_node.idx is not None:
            name = '%s_%d' % (name, des_var_node.idx)
        return cls(name, bounds=des_var_node.bounds, options=des_var_node.options, node=des_var_node,
                   conditionally_active=conditionally_active)

    @classmethod
    def from_choice_node(cls, choice_node: ChoiceNode, options: list, name: str = None,
                         existing_names: set = None, conditionally_active=False) -> 'DesVar':
        if name is None:
            name = choice_node.decision_id

        if existing_names is not None:
            name_base = name
            i = 2
            while name in existing_names:
                name = f'{name_base}_{i}'
                i += 1

        return cls(name, options=options, node=choice_node, conditionally_active=conditionally_active)

    @property
    def name(self) -> str:
        """Design variable name"""
        return self._name

    @property
    def is_discrete(self) -> bool:
        """Whether the design variable is discrete or continuous"""
        return self._opts is not None

    @property
    def options(self) -> Optional[list]:
        """List of options (if discrete)"""
        return self._opts

    @property
    def n_opts(self) -> Optional[int]:
        """Number of options (if discrete)"""
        return len(self._opts) if self._opts is not None else None

    @property
    def bounds(self) -> Optional[Tuple[float, float]]:
        """Bounds (if continuous)"""
        return self._bounds

    @property
    def node(self) -> Optional[Union[DesignVariableNode, ChoiceNode]]:
        """Associated choice node"""
        return self._node

    def rand(self):
        """Generate a random value"""
        if self.is_discrete:
            return random.randint(0, self.n_opts-1)
        return random.uniform(*self.bounds)

    def __str__(self):
        if self.is_discrete:
            return f'DV: {self.name} [{self.n_opts} opts]'
        return f'DV: {self.name} [{self.bounds[0]:.2f}..{self.bounds[1]:.2f}]'

    def __repr__(self):
        return str(self)


class Direction(enum.Enum):
    MIN = -1
    MAX = 1
    LTE = -1
    GTE = 1


class Objective:
    """Class representing an objective."""

    def __init__(self, name: str, direction=Direction.MIN, node: MetricNode = None):
        self._name = name
        self._dir = direction
        self._node = node

    @classmethod
    def from_metric_node(cls, metric_node: MetricNode) -> 'Objective':
        name = metric_node.name
        if metric_node.dir is None:
            raise ValueError(f'Metric node has no direction specified: {name}')

        if metric_node.idx is not None:
            name = f'{name}_{metric_node.idx}'

        direction = Direction.MIN if metric_node.dir <= 0 else Direction.MAX
        return cls(name, direction, node=metric_node)

    @property
    def name(self) -> str:
        """Objective name"""
        return self._name

    @property
    def dir(self) -> Direction:
        return self._dir

    @property
    def sign(self) -> int:
        """Objective direction (-1 = minimization, 1 = maximization)"""
        return self._dir.value

    @property
    def node(self) -> Optional[MetricNode]:
        """Associated metric node"""
        return self._node

    def __str__(self):
        return f'OBJ: {self.name} [{"min" if self.sign < 0 else "max"}]'

    def __repr__(self):
        return str(self)


class Constraint:
    """Class representing an inequality constraint. The direction specifies the side which is considered feasible."""

    def __init__(self, name: str, ref: float = 0., direction=Direction.LTE, node: MetricNode = None):
        self._name = name
        self._ref = ref
        self._dir = direction
        self._node = node

    @classmethod
    def from_metric_node(cls, metric_node: MetricNode) -> 'Constraint':
        name = metric_node.name
        if metric_node.dir is None:
            raise ValueError(f'Metric node has no direction specified: {name}')
        if metric_node.ref is None:
            raise ValueError(f'Metric node has no reference value specified: {name}')

        if metric_node.idx is not None:
            name = f'{name}_{metric_node.idx}'

        direction = Direction.LTE if metric_node.dir <= 0 else Direction.GTE
        return cls(name, metric_node.ref, direction, node=metric_node)

    @property
    def name(self) -> str:
        """Objective name"""
        return self._name

    @property
    def ref(self) -> float:
        """Reference value (the value should be above/below this value)"""
        return self._ref

    @property
    def dir(self) -> Direction:
        return self._dir

    @property
    def sign(self) -> int:
        """Constraint direction (-1 = lower than or equal, 1 = greater than or equal)"""
        return self._dir.value

    @property
    def node(self) -> Optional[MetricNode]:
        """Associated metric node"""
        return self._node

    def __str__(self):
        return f'CON: {self.name} [{"<=" if self.sign < 0 else ">="} {self.ref:.2f}]'

    def __repr__(self):
        return str(self)



import string
import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['EnumRecursiveEncoder']


class EnumRecursiveEncoder(QuasiLazyEncoder):
    """Defines design variables where each has a fixed nr of options to cover all possible assignment patterns."""

    _str_lookup = {char: i for i, char in enumerate(string.digits+string.ascii_uppercase)}

    def __init__(self, *args, n_divide=2, **kwargs):
        self.n_divide = max(2, n_divide)
        super().__init__(*args, **kwargs)
        self._dv_last = {}
        self._dv_inactive_key = {}

    def _encode_prepare(self):
        super()._encode_prepare()
        self._dv_last = {}
        self._dv_inactive_key = {}

    def _encode_matrix(self, matrix: np.ndarray, existence: NodeExistence) -> List[DiscreteDV]:
        n_mat = matrix.shape[0]
        if n_mat <= 1:
            return []

        n = self.n_divide
        n_var = int(np.ceil(np.log(n_mat)/np.log(n)))

        # Get design vector values that lead to inactive variables (due to nr cutoff)
        dv_last = np.array(self.base_repr_int(n_mat-1, n))
        i_inactive = np.where(dv_last == 0)[0]
        if len(i_inactive) > 0:
            left_side_values = dv_last[:i_inactive[-1]+1].copy()
            self._dv_inactive_key[existence] = (i_inactive, left_side_values)
            dv_last[i_inactive] = X_INACTIVE_VALUE
        self._dv_last[existence] = dv_last

        n_opts = np.ones((n_var,), dtype=int)*n
        n_opts[0] = dv_last[0]+1
        return [DiscreteDV(n_opts=n_opt, conditionally_active=dv_last[i] == X_INACTIVE_VALUE)
                for i, n_opt in enumerate(n_opts)]

    def _decode_matrix(self, vector: DesignVector, matrix: np.ndarray, existence: NodeExistence) \
            -> Optional[Tuple[DesignVector, np.ndarray]]:
        if len(vector) == 0:
            return vector, matrix[0, :, :]

        i_mat = np.sum((self.n_divide**np.arange(len(vector)))*vector[::-1])
        if i_mat >= matrix.shape[0]:
            return self._dv_last[existence], matrix[-1, :, :]

        vectors = np.array([vector])
        self._set_inactive(vectors, existence)
        vector = vectors[0, :]

        return vector, matrix[i_mat, :, :]

    @classmethod
    def base_repr_int(cls, value: int, base: int) -> List[int]:
        return [cls._str_lookup[char] for char in (np.binary_repr(value) if base == 2 else np.base_repr(value, base))]

    def _do_generate_random_dv_mat(self, n: int, existence: NodeExistence, matrix: np.ndarray,
                                   design_vars: List[DiscreteDV]) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        if n < matrix.shape[0]:
            i_selected = np.random.choice(matrix.shape[0], n, replace=False)
        else:
            i_selected = np.arange(matrix.shape[0])

        design_vectors = self._encode_design_vectors(i_selected, len(design_vars), existence)
        matrices = matrix[i_selected, :, :]
        return design_vectors, matrices

    def _do_get_all_design_vectors(self, existence: NodeExistence, matrix: np.ndarray, design_vars: List[DiscreteDV]) \
            -> np.ndarray:
        design_vectors = self._encode_design_vectors(np.arange(matrix.shape[0]), len(design_vars), existence)
        return design_vectors

    def _encode_design_vectors(self, i_selected, n_dv: int, existence: NodeExistence) -> np.ndarray:
        design_vectors = np.zeros((len(i_selected), n_dv), dtype=int)
        for i_dv, i_sel in enumerate(i_selected):
            base_repr = self.base_repr_int(i_sel, self.n_divide)
            design_vectors[i_dv, -len(base_repr):] = base_repr

        self._set_inactive(design_vectors, existence)
        return design_vectors

    def _set_inactive(self, design_vectors: np.ndarray, existence: NodeExistence):
        dv_inactive_key = self._dv_inactive_key.get(existence)
        if dv_inactive_key is not None:
            i_inactive, left_side_values = dv_inactive_key
            i_dv_inactive = np.all(design_vectors[:, :len(left_side_values)] == left_side_values, axis=1)
            for ix in i_inactive:
                design_vectors[i_dv_inactive, ix] = X_INACTIVE_VALUE

    def __repr__(self):
        return f'{self.__class__.__name__}({self._imputer!r})'

    def __str__(self):
        return f'Enum Rec {self.n_divide} + {self._imputer!s}'



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import math
from typing import *
from adsg_core.graph.adsg_basic import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.optimization.evaluator import DSGEvaluator

__all__ = ['ApolloEvaluator']


class ApolloDecisionVar(NamedNode):

    def __init__(self, decision: str, value):
        self.decision = decision
        self.value = value
        super().__init__(f'{decision}:{value!s}')

    def get_export_title(self) -> str:
        return f'{self.decision} = {self.value}'


class ApolloEvaluator(DSGEvaluator):
    """
    Class representing the Apollo system architecture design problem, as presented by:
    Simmons 2008: A Framework for Decision Support in Systems Architecting

    The problem involves selection of the Apollo moon mission and vehicle architectures:
    - Mass (minimization) and success probability (maximization) are optimized
    - Mission architecture is selected:
      - Earth Orbit Rendezvous (yes/no)
      - Earth launch (orbit/direct)
      - Lunar Orbit Rendezvous (yes/no)
      - Moon arrival (orbit/direct)
      - Moon departure (orbit/direct)
    - Vehicle architecture is selected:
      - Command module crew (2/3)
      - Lunar module crew (0/1/2/3)
      - Service module fuel (cryogenic/storable)
      - Lunar module fuel (NA/cryogenic/storable)
    """

    risk_table = {
        'EOR': {False: .98, True: .95},
        'earthLaunch': {'orbit': .99, 'direct': .9},
        'LOR': {False: 1., True: .95},
        'moonArrival': {'orbit': .99, 'direct': .95},
        'moonDeparture': {'orbit': .9, 'direct': .9},
        'cmCrew': {2: 1., 3: 1.},
        'lmCrew': {0: 1., 1: .9, 2: 1., 3: 1.},
        'smFuel': {'cryogenic': .95, 'storable': 1.},
        'lmFuel': {'NA': 1., 'cryogenic': .9025, 'storable': 1.},
    }

    def __init__(self, objective: int = None):
        super().__init__(self.get_adsg(objective=objective))

    @staticmethod
    def get_adsg(objective: int = None):
        metric_nodes = []
        if objective is None or objective == 0:
            metric_nodes.append(MetricNode('mass', direction=-1, type_=MetricType.OBJECTIVE))
        if objective is None or objective == 1:
            metric_nodes.append(MetricNode('success', direction=1, type_=MetricType.OBJECTIVE))
        if len(metric_nodes) == 0:
            raise ValueError('No objectives specified!')

        dsg = BasicDSG()
        start_nodes = set()

        # EARTH LAUNCH DECISION #
        # If we choose EOR (Earth orbit rendezvous), we have to launch to orbit
        # We represent this constraint by first adding the EOR yes/no decision,
        # and then adding launch decisions after the no-EOR node
        earth_launch = NamedNode('earthLaunch')
        dsg.add_edges([(earth_launch, mn) for mn in metric_nodes])
        start_nodes.add(earth_launch)

        eor_yes = ApolloDecisionVar('EOR', True)
        eor_no = ApolloDecisionVar('EOR', False)
        earth_launch_orbit = ApolloDecisionVar('earthLaunch', 'orbit')
        earth_launch_direct = ApolloDecisionVar('earthLaunch', 'direct')

        dsg.add_selection_choice('EOR', earth_launch, [eor_no, eor_yes])
        dsg.add_selection_choice('earthLaunch', eor_no, [earth_launch_orbit, earth_launch_direct])
        dsg.add_edge(eor_yes, earth_launch_orbit)

        # MOON ARRIVAL/DEPARTURE DECISIONS #
        # If we choose LOR, we have to arrive and depart by orbit
        # We represent this constraint by first adding the LOR yes/no decision, and then adding arrival/departure
        # decisions after the no-LOR node
        moon_arr_dep = NamedNode('moonArrDep')
        start_nodes.add(moon_arr_dep)

        lor_yes = ApolloDecisionVar('LOR', True)
        lor_no = ApolloDecisionVar('LOR', False)
        moon_arr_orbit = ApolloDecisionVar('moonArrival', 'orbit')
        moon_arr_direct = ApolloDecisionVar('moonArrival', 'direct')
        moon_dep_orbit = ApolloDecisionVar('moonDeparture', 'orbit')
        moon_dep_direct = ApolloDecisionVar('moonDeparture', 'direct')

        dsg.add_selection_choice('moonLOR', moon_arr_dep, [lor_no, lor_yes])
        dsg.add_selection_choice('moonArr', lor_no, [moon_arr_orbit, moon_arr_direct])
        dsg.add_selection_choice('moonDep', lor_no, [moon_dep_orbit, moon_dep_direct])
        dsg.add_edges([
            (lor_yes, moon_arr_orbit),
            (lor_yes, moon_dep_orbit),
        ])

        # CREW DECISION #
        # Select either 2 or 3 crew in the command module (aka the mission)
        crew = NamedNode('crew')
        start_nodes.add(crew)

        crew_2 = ApolloDecisionVar('cmCrew', 2)
        crew_3 = ApolloDecisionVar('cmCrew', 3)
        dsg.add_selection_choice('crew', crew, [crew_2, crew_3])

        # SERVICE MODULE FUEL DECISION #
        # Select fuel type for the service module (connected to the command module)
        service_fuel = NamedNode('smFuel')
        start_nodes.add(service_fuel)

        sm_fuel_cryogenic = ApolloDecisionVar('smFuel', 'cryogenic')
        sm_fuel_storable = ApolloDecisionVar('smFuel', 'storable')
        dsg.add_selection_choice('smFuel', service_fuel, [sm_fuel_cryogenic, sm_fuel_storable])

        # LUNAR MODULE DECISIONS #
        # Whether there is a lunar module is determined by whether there is a LOR
        # If yes, we select the nr of crew members and the fuel type
        # If no, we select 0 crew members and "NA" fuel type
        dsg.add_edge(lor_no, ApolloDecisionVar('lmCrew', 0))
        lm_crew = [ApolloDecisionVar('lmCrew', n) for n in [1, 2, 3]]
        dsg.add_selection_choice('lmCrew', lor_yes, lm_crew)

        dsg.add_edge(lor_no, ApolloDecisionVar('lmFuel', 'NA'))
        dsg.add_selection_choice('lmFuel', lor_yes, [
            ApolloDecisionVar('lmFuel', 'cryogenic'), ApolloDecisionVar('lmFuel', 'storable')])

        # Constrain that if we select 2 crew members, we cannot select 3 lunar module crew members
        dsg.add_incompatibility_constraint([crew_2, lm_crew[-1]])

        # Set start nodes
        dsg = dsg.set_start_nodes(start_nodes)

        return dsg

    def _evaluate(self, dsg: DSGType, metric_nodes: List[MetricNode]) -> Dict[MetricNode, float]:
        # Get decision values from selected ApolloDecisionVar nodes
        decision_values = {
            'EOR': False,
            'earthLaunch': 'direct',
            'LOR': False,
            'moonArrival': 'direct',
            'moonDeparture': 'direct',
            'lmCrew': 0,
            'lmFuel': 'NA',
        }
        dv_node: ApolloDecisionVar
        for dv_node in dsg.get_nodes_by_type(ApolloDecisionVar):
            decision_values[dv_node.decision] = dv_node.value

        # Check that no constraint is violated
        self.is_possible(decision_values)

        # Calculate metrics: the outputs of this function should be a dict mapping requested metric nodes to values
        results = {}
        for metric_node in metric_nodes:
            if metric_node.name == 'mass':
                results[metric_node] = self.eval_mass(decision_values)
            elif metric_node.name == 'success':
                results[metric_node] = self.eval_success_probability(decision_values)

        return results

    @staticmethod
    def is_possible(dv):  # Table 4-3 (p.104)

        if dv['EOR'] and dv['earthLaunch'] != 'orbit':  # EORconstraint
            raise RuntimeError('EOR constraint violated!')

        if dv['LOR'] and (dv['moonArrival'] != 'orbit' or dv['moonDeparture'] != 'orbit'):  # LORconstraint, moonLeaving
            raise RuntimeError('LOR constraint violated!')

        if dv['cmCrew'] < dv['lmCrew']:  # lmcmcrew
            raise RuntimeError('Crew constraint violated!')

        if dv['LOR'] == (dv['lmCrew'] == 0):  # lmexists
            raise RuntimeError('Lunar module constraint violated!')

        if dv['LOR'] == (dv['lmFuel'] == 'NA'):  # lmFuelConstraint
            raise RuntimeError('Fuel constraint violated!')

    @classmethod
    def eval_success_probability(cls, dv):
        prob = 1.
        for key, value in dv.items():

            # Special case for cryogenic service module fuel
            if key == 'smFuel' and value == 'cryogenic':
                n_burn = 2. if dv['LOR'] else 4.
                prob *= math.pow(cls.risk_table[key][value], n_burn)
                continue

            try:
                prob *= cls.risk_table[key][value]
            except KeyError:
                raise KeyError('Risk table key not found: %s, %r' % (key, value))
        return prob

    @staticmethod
    def eval_mass(dv):
        """
        Based on the Jython script provided in Appendix A of:
        Simmons2008 A Framework for Decision Support in Systems Architecting

        :param dv:
        :rtype: float
        """

        d_veod = 11100
        d_vmoe = 3840
        d_vdom = 6798
        d_vmoa = 7468
        dv_mod = 3661
        g = 9.8 * 100 / 30
        i_sp1 = 315
        i_sp2 = 425
        f = 0.08

        def cm_mass(cm_crew):
            if cm_crew == 2:
                return 8000
            if cm_crew == 3:
                return 11000
            return 0

        def lm_mass(lm_crew):
            if lm_crew == 1:
                return 3000
            if lm_crew == 2:
                return 4000
            if lm_crew == 3:
                return 5000
            return 0

        def sm_4_mass(m_cm, i_sp):
            return stage_mass(dv_mod, i_sp, m_cm)

        def lm_2_mass(lor, m_lm, i_sp):
            if not lor:
                return 0

            return stage_mass(d_vmoa, i_sp, m_lm)

        def lm_1_mass(lor, m_lm2, m_lm, i_sp):
            if not lor:
                return 0

            m_payload = m_lm2 + m_lm
            return stage_mass(d_vdom, i_sp, m_payload)

        def sm_3_mass(lor, m_sm4, m_cm, i_sp):
            if lor:
                return 0

            m_payload = m_sm4 + m_cm
            return stage_mass(d_vmoa, i_sp, m_payload)

        def sm_2_mass(lor, m_sm3, m_sm4, m_cm, i_sp):
            if lor:
                return 0

            m_payload = m_sm3 + m_sm4 + m_cm
            return stage_mass(d_vdom, i_sp, m_payload)

        def sm_1_mass(m_lm_total, m_sm2, m_sm3, m_sm4, m_cm, i_sp):
            m_payload = m_lm_total + m_sm2 + m_sm3 + m_sm4 + m_cm
            return stage_mass(d_vmoe, i_sp, m_payload)

        def s_4_mass(m_payload):
            return stage_mass(d_veod, i_sp2, m_payload)

        def r_calc(delta_v, i_sp):
            x = (delta_v / (i_sp * g))
            ratio = math.exp(-x)
            return ratio

        def rt2p(delta_v, i_sp):
            r = r_calc(delta_v, i_sp)
            return 1 / (r - f + (f * r))

        def stage_mass(delta_v, i_sp, m_payload):
            total_payload_ratio = rt2p(delta_v, i_sp)
            return total_payload_ratio*m_payload - m_payload

        def calc_i_sp(fuel_type):
            return i_sp2 if fuel_type == 'cryogenic' else i_sp1

        def im_leo(lor, cm_crew, lm_crew, sm_fuel_type, lm_fuel_type):
            m_cm = cm_mass(cm_crew)
            m_sm4 = sm_4_mass(m_cm, calc_i_sp(sm_fuel_type))

            m_lm = lm_mass(lm_crew)
            m_lm2 = lm_2_mass(lor, m_lm, calc_i_sp(lm_fuel_type))
            m_lm1 = lm_1_mass(lor, m_lm2, m_lm, calc_i_sp(lm_fuel_type))
            m_lm_total = sum([m_lm1, m_lm2, m_lm])

            m_sm3 = sm_3_mass(lor, m_sm4, m_cm, calc_i_sp(sm_fuel_type))
            m_sm2 = sm_2_mass(lor, m_sm3, m_sm4, m_cm, calc_i_sp(sm_fuel_type))
            m_sm1 = sm_1_mass(m_lm_total, m_sm2, m_sm3, m_sm4, m_cm, calc_i_sp(sm_fuel_type))

            m_total = sum([m_sm1, m_sm2, m_sm3, m_sm4])

            return sum([
                s_4_mass(sum([m_total, m_lm_total, m_cm])),
                m_total,
                m_lm_total,
                m_cm,
            ])

        return im_leo(dv['LOR'], dv['cmCrew'], dv['lmCrew'], dv['smFuel'], dv['lmFuel'])


if __name__ == '__main__':
    # After export, visualize contents using https://viz-js.com/
    ApolloEvaluator.get_adsg().export_dot('apollo.dot')

    evaluator = ApolloEvaluator()
    for _ in range(10):
        graph, _, _ = evaluator.get_graph(evaluator.get_random_design_vector())
        evaluator.evaluate(graph)



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import *
import networkx as nx
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *

__all__ = ['get_nodes_by_type', 'get_in_degree', 'get_out_degree', 'get_choice_nodes', 'get_derived_edges_for_edge',
           'get_derived_edges_for_node', 'get_deriving_in_edges', 'check_derives', 'has_conditional_existence',
           'get_confirmed_edges_for_node', 'get_unconnected_connectors', 'traverse_until_choice_nodes',
           'get_non_confirmed_nodes', 'get_nodes_by_subtype', 'iter_in_edges_cached', 'iter_out_edges_cached']


T = TypeVar('T')


def get_nodes_by_type(graph: nx.MultiDiGraph, type_: Type[T]) -> List[T]:
    return [node for node in graph.nodes if type(node) == type_]


def get_nodes_by_subtype(graph: nx.MultiDiGraph, type_: Type[T]) -> List[T]:
    return [node for node in graph.nodes if isinstance(node, type_)]


def get_in_degree(graph: nx.MultiDiGraph, node: DSGNode, edge_type: EdgeType = None) -> int:
    if edge_type is None:
        return graph.in_degree(node)

    n_in = 0
    for edge in iter_in_edges(graph, node):
        if get_edge_type(edge) == edge_type:
            n_in += 1
    return n_in


def get_out_degree(graph: nx.MultiDiGraph, node: DSGNode, edge_type: EdgeType = None) -> int:
    if edge_type is None:
        return graph.out_degree(node)

    n_out = 0
    for edge in iter_out_edges(graph, node):
        if get_edge_type(edge) == edge_type:
            n_out += 1
    return n_out


def get_choice_nodes(graph: nx.MultiDiGraph) -> List[ChoiceNode]:
    return get_nodes_by_type(graph, SelectionChoiceNode)+get_nodes_by_type(graph, ConnectionChoiceNode)


def get_derived_edges_for_edge(graph, edge, start_nodes, removed_edges=None, removed_nodes=None,
                               traversed: Set[DSGNode] = None, cache=None):
    return get_derived_edges_for_node(graph, edge[0], start_nodes, removed_edges=removed_edges,
                                      removed_nodes=removed_nodes, derived_edges={edge}, traversed=traversed,
                                      cache=cache)


def get_derived_edges_for_node(
        graph: nx.MultiDiGraph,
        node: DSGNode,
        start_nodes: Optional[Set[DSGNode]],
        removed_edges: Set[Union[EdgeTuple, Tuple[DSGNode, DSGNode]]] = None,
        removed_nodes: Set[DSGNode] = None,
        derived_edges: Set[EdgeTuple] = None,
        traversed: Set[DSGNode] = None,
        cache=None,
) -> Tuple[Set[EdgeTuple], Set[DSGNode]]:
    """
    Get the edges and nodes for a given node that are derived ONLY by this node.

    In general, from the starting node, all successor nodes/edges are derived if these nodes do not have additional
    in-edges not stemming from this origin node. In-edges that are in the removed_edges array are not taken into
    account. In addition, for connector nodes, in-edges stemming from other connector nodes are also not taken into
    account (because these do not serve as a "reason" for the connector node to exist).
    """

    derived_nodes = set()
    if removed_edges is None:
        removed_edges = set()
    if removed_nodes is None:
        removed_nodes = set()
    if traversed is None:
        traversed = {node}
    removed_derived_nodes = set(removed_nodes)

    # All out-edges are derived edges (if not given)
    if derived_edges is None:
        derived_edge_types = [EdgeType.DERIVES, EdgeType.CONNECTS]
        derived_edges = {edge for edge in iter_out_edges_cached(graph, node, cache=cache)
                         if get_edge_type(edge) in derived_edge_types}
    removed_derived_edges = removed_edges | derived_edges

    # For every out-edge, continue determining the derived edges
    for edge in list(derived_edges):
        target_node = edge[1]
        edge_type = get_edge_type(edge, default=EdgeType.DERIVES)

        # If we already checked this node, do not continue checking it
        if target_node in traversed:
            continue

        # If the target node is contained in the set of start nodes, regard it as an independent node
        if start_nodes is not None and target_node in start_nodes:
            continue

        # Get the derived edges and nodes of the target node
        target_derived_edges, target_derived_nodes = \
            get_derived_edges_for_node(graph, target_node, start_nodes, removed_edges=removed_derived_edges,
                                       removed_nodes=removed_derived_nodes | {target_node},
                                       traversed=traversed | {target_node}, cache=cache)

        # Get the in-edges deriving the target node, not taking the derived edges into account
        deriving_in_edges = get_deriving_in_edges(
            graph, target_node, removed_edges=removed_derived_edges, removed_nodes=removed_derived_nodes,
            edge_type=edge_type, cache=cache)

        # Remove any in-edge that is also derived by the target node
        deriving_in_edges -= target_derived_edges

        # If there are any remaining in-edges, it means that the target node could be derived by other means than
        # the origin node, so we do not consider it as a derived node
        if len(deriving_in_edges) > 0:
            continue

        # The target node is derived ONLY by the origin node, so we also consider its derived edges and nodes
        derived_edges |= target_derived_edges
        derived_nodes |= {target_node} | target_derived_nodes
        removed_derived_nodes |= {target_node} | target_derived_nodes

    return derived_edges, derived_nodes


def get_deriving_in_edges(graph: nx.MultiDiGraph, node: DSGNode, removed_edges: Iterable[tuple] = None,
                          removed_nodes: Iterable[DSGNode] = None, edge_type: EdgeType = None, cache=None) \
        -> Set[EdgeTuple]:
    """
    Returns in-edges of a node that are not in the removed_edges set or (indirectly) stemming from needed port
    nodes.
    """

    # Get non-removed in-edges
    if removed_edges is None:
        removed_edges = set()
    if removed_nodes is None:
        removed_nodes = set()

    # removing_nodes = frozenset({edge[0] for edge in removed_edges if edge[1] is node} | removed_nodes)

    # # Load cache if available
    # der_in_cache = cache_key = None
    # if cache is not None:
    #     cache['der_in_edges'] = der_in_cache = cache.get('der_in_edges', {})
    #
    #     # Load from cache if available
    #     cache_key = hash((node, edge_type, removing_nodes))
    #     if cache_key in der_in_cache:
    #         return der_in_cache[cache_key].copy()

    in_edges = set()
    deriving_edge_types = {EdgeType.DERIVES, edge_type}
    for edge in iter_in_edges_cached(graph, node, cache=cache):
        # Check if edge is in removed edges
        # if edge[0] in removing_nodes:
        if edge[0] in removed_nodes or edge in removed_edges or (edge[0], edge[1]) in removed_edges:
            continue

        # Check edge type
        if get_edge_type(edge, default=EdgeType.DERIVES) in deriving_edge_types:
            in_edges.add(edge)

    # if der_in_cache is not None:
    #     der_in_cache[cache_key] = in_edges.copy()

    return in_edges


def check_derives(graph: nx.MultiDiGraph, source_node: DSGNode, target_node: DSGNode, connects=False) -> bool:

    allowed_types = {EdgeType.DERIVES}
    if connects:
        allowed_types.add(EdgeType.CONNECTS)

    for edge in iter_in_edges(graph, target_node):
        edge_type = get_edge_type(edge)
        if edge_type not in allowed_types:
            continue

        if edge[0] == source_node:
            return True

        if check_derives(graph, source_node, edge[0], connects=connects):
            return True
    return False


def has_conditional_existence(graph: nx.MultiDiGraph, start_nodes: Set[DSGNode], target_node: DSGNode,
                              base_nodes: List[DSGNode] = None) -> int:
    """
    Returns whether a node exists conditionally or not. Conditional existence means that due to some decision, there
    is a possibility that the node will not exist when taking this decision.

    There are three return values:
    - 0: Unconditional existence: target node will always exist
    - 1: Conditional existence: based on some intermediate decision, it can be possible that the target node does
                                not exist
    - 2: Unconditional non-existence: target node will never exist if base nodes exist (i.e. none of the base nodes
                                      derive the target node)
    """

    derives_cache = {}
    derives_traversed = {}

    # Function to check whether some node derives another node
    def _derives(base_node, check_node, start_check_node=None, removed_nodes=None, decision_cutoff=False):
        if base_node == check_node:
            return True
        removed_nodes_key = tuple(sorted([hash(n) for n in removed_nodes or []]))
        cache_key = (base_node, check_node, removed_nodes_key, decision_cutoff)
        if cache_key in derives_cache:
            return derives_cache[cache_key]

        if start_check_node is None:
            start_check_node = check_node
        traversed_key = (base_node, start_check_node, removed_nodes_key, decision_cutoff)
        if traversed_key not in derives_traversed:
            derives_traversed[traversed_key] = set()
        derives_traversed[traversed_key].add(check_node)

        # Loop over in-edges
        does_derive = False
        for edge in iter_in_edges(graph, check_node):
            if get_edge_type(edge) != EdgeType.DERIVES:
                continue

            derives_cache[edge[0], edge[1], removed_nodes_key, decision_cutoff] = True
            if edge[0] == base_node:
                return True

            # Prevent infinite recursion
            if edge[0] in derives_traversed[traversed_key]:
                continue

            # If the incoming node is removed, the link is broken
            if removed_nodes is not None and edge[0] in removed_nodes:
                continue

            # Check if we should continue at decision nodes
            if decision_cutoff and isinstance(edge[0], ChoiceNode):
                continue

            # Walk back
            if _derives(base_node, edge[0], start_check_node=start_check_node, removed_nodes=removed_nodes,
                        decision_cutoff=decision_cutoff):
                does_derive = True
                break

        derives_cache[cache_key] = does_derive
        return does_derive

    # Get nodes to check conditional existence from
    if base_nodes is None:
        base_nodes = start_nodes
    else:
        base_nodes = set(base_nodes)

        # If we are comparing to some custom target nodes, we do not have conditional existence if the target node
        # derives one of the base nodes
        for check_base_node in base_nodes:
            if _derives(target_node, check_base_node):
                return False

    if not base_nodes or len(base_nodes) == 0:
        return False

    traversed = {}
    in_walk_back = set()

    # Function to check whether some base node always derives some other node
    def _maybe_derived_from_start_node(base_node, node=None):
        # If node is not given, we want to check for the base node
        if node is None:
            node = base_node

        # If the base node is one of the ext func nodes, the node is indeed always derived from an ext func node
        if base_node in base_nodes:
            return False

        # Check cache (also to prevent infinite loops)
        if base_node in traversed:
            return traversed[base_node]

        in_walk_back.add(base_node)

        # If we are at a decision node, check whether all of the options derive the target node
        is_choice = isinstance(base_node, ChoiceNode)
        is_conditional = False
        if is_choice and node != base_node:
            # Loop over out edges (the decision-options)
            option_nodes = [edge[1] for edge in iter_out_edges(graph, base_node)
                            if get_edge_type(edge) == EdgeType.DERIVES]

            for option_node in option_nodes:
                removed_nodes = set(option_nodes)
                strictly_derives = _derives(option_node, node, decision_cutoff=True)
                if not strictly_derives:
                    removed_nodes -= {option_node}

                # If a specific option does not derive the target node, not all paths lead to the target node and
                # therefore the target node is also not always derived from an ext func node
                if not _derives(option_node, node, removed_nodes=removed_nodes):
                    is_conditional = True
                    break

        # Walk back
        derivation_status = 2
        for edge in iter_in_edges(graph, base_node):
            if get_edge_type(edge) != EdgeType.DERIVES:
                continue

            if edge[0] in in_walk_back:
                continue

            # If any of the previous nodes is always derived from an external function node, we are too
            maybe_derived_status = _maybe_derived_from_start_node(edge[0], node=node)
            if not maybe_derived_status:
                traversed[base_node] = is_conditional
                return is_conditional

            # If any of the previous node has conditional existence, we also have if we don't find an unconditional node
            if maybe_derived_status == 1:
                derivation_status = 1

        # If none of the previous nodes are derived from an external function node, we are neither
        traversed[base_node] = derivation_status
        return derivation_status

    return _maybe_derived_from_start_node(target_node)


def get_confirmed_edges_for_node(graph: nx.MultiDiGraph, node: DSGNode, include_choice=False,
                                 _traversed=None, _traversed_to_update=None, _walked_hit=None,
                                 cache=None) -> Set[EdgeTuple]:

    # Initialize temp storage if needed
    # _traversed: store for each traversed node which edges were marked as confirmed
    # _traversed_to_update: node pairs (src, tgt) where the edges of the tgt should be the same as edges of the src
    # _walked_hit: temp cache to remember which nodes are part of a loop when encountering a previously-traversed node
    is_request_start = False
    if _traversed is None:
        _traversed = {node: set()}
        is_request_start = True
    if _traversed_to_update is None:
        _traversed_to_update = []
    if _walked_hit is None:
        _walked_hit = {}

    # Load cache if available
    conf_edges_cache = None
    if cache is not None:
        cache['conf_edge'] = conf_edges_cache = cache.get('conf_edge', {})

        # Load from cache if available
        if node in conf_edges_cache:
            return conf_edges_cache[node].copy()

    # Loop over outgoing edges
    confirmed_edges = set()
    for out_edge in iter_out_edges(graph, node):
        if get_edge_type(out_edge) == EdgeType.INCOMPATIBILITY:
            continue

        # Stop at choice nodes
        if isinstance(out_edge[1], ChoiceNode):
            # Mark edge as confirmed if we also want to return choice nodes
            if include_choice:
                confirmed_edges.add(out_edge)
            continue

        # Check if already walked
        if out_edge[1] in _traversed:
            # Remember where we obtained this hit
            if node not in _walked_hit:
                _walked_hit[node] = []
            _walked_hit[node].append(out_edge[1])

            # Mark the edge as confirmed
            confirmed_edges.add(out_edge)
            continue

        if out_edge[1] not in _traversed:
            _traversed[out_edge[1]] = set()

        # Recursively walk the graph
        confirmed_edges.add(out_edge)
        child_confirmed_edges = get_confirmed_edges_for_node(
            graph, out_edge[1], include_choice=include_choice,
            _traversed=_traversed, _traversed_to_update=_traversed_to_update, _walked_hit=_walked_hit, cache=cache)

        confirmed_edges |= child_confirmed_edges

        # If the target node lead to a (downstream) "walked hit"
        if out_edge[1] in _walked_hit:

            # Extend the update list to include the current edge target
            hit_nodes = _walked_hit[out_edge[1]]
            for tgt_nodes in hit_nodes:
                _traversed_to_update.append((out_edge[1], tgt_nodes))

            # Propagate "walked hit" nodes upstream
            _walked_hit[out_edge[1]] = hit_nodes = [nd for nd in hit_nodes if nd != node]
            if len(hit_nodes) > 0:
                if node not in _walked_hit:
                    _walked_hit[node] = []
                _walked_hit[node] += hit_nodes

    # Update the traversed edges for this node
    _traversed[node] = confirmed_edges

    # Update traversed edges for nodes part of a loop
    for tgt_node, src_node in _traversed_to_update:
        _traversed[tgt_node].update(_traversed[src_node])

    # Update cache only if this was the originally-requested start node
    if conf_edges_cache is not None and is_request_start:
        for start_node, edges in _traversed.items():
            if start_node not in conf_edges_cache:
                conf_edges_cache[start_node] = edges

    return confirmed_edges


def get_unconnected_connectors(graph: nx.MultiDiGraph, start_nodes: Set[DSGNode], stop_at_one: bool = False)\
        -> List[ConnectorNode]:

    checked = {}
    unconnected_connectors = []
    for connector_node in get_nodes_by_subtype(graph, ConnectorNode):
        base_conn_node = connector_node
        for edge in iter_out_edges(graph, connector_node, edge_type=EdgeType.DERIVES):
            if isinstance(edge[1], ConnectorDegreeGroupingNode):
                base_conn_node = edge[1]
                break

        is_out_conn = True
        for edge in iter_in_edges(graph, base_conn_node):
            if get_edge_type(edge) == EdgeType.CONNECTS:
                is_out_conn = False
                break

        next_node = None
        for nested_edge in (iter_out_edges(graph, base_conn_node, edge_type=EdgeType.CONNECTS)
                            if is_out_conn else iter_in_edges(graph, base_conn_node, edge_type=EdgeType.CONNECTS)):
            next_node = nested_edge[1] if is_out_conn else nested_edge[0]
            break

        # next_node = self.next(connector_node, include_decisions=True)
        # if len(next_node) == 1 and isinstance(next_node[0], ConnectorDegreeGroupingNode):
        #     base_conn_node = next_node[0]
        #     next_node = self.next(next_node[0], include_decisions=True)

        if base_conn_node in checked:
            if checked[base_conn_node] and base_conn_node is not connector_node:
                unconnected_connectors.append(connector_node)
            continue

        # # If there is still a decision to be taken, in principle everything should be fine
        if next_node is not None and isinstance(next_node, ConnectionChoiceNode):
            conn_deg = get_out_degree(graph, next_node, edge_type=EdgeType.CONNECTS) \
                if is_out_conn else get_in_degree(graph, next_node, edge_type=EdgeType.CONNECTS)

            if not base_conn_node.is_valid(0) and conn_deg == 0 and \
                    not has_conditional_existence(graph, start_nodes, base_conn_node):
                unconnected_connectors.append(connector_node)
                if stop_at_one:
                    return unconnected_connectors
                checked[base_conn_node] = True
                continue

        else:
            conn_deg = get_out_degree(graph, base_conn_node, edge_type=EdgeType.CONNECTS) \
                if is_out_conn else get_in_degree(graph, base_conn_node, edge_type=EdgeType.CONNECTS)
            if not base_conn_node.is_valid(conn_deg):
                unconnected_connectors.append(connector_node)
                if stop_at_one:
                    return unconnected_connectors
                checked[base_conn_node] = True
                continue

        checked[base_conn_node] = False

    return unconnected_connectors


def traverse_until_choice_nodes(graph: nx.MultiDiGraph, start_nodes: Set[DSGNode], traversed: set = None) \
        -> Tuple[Set[DSGNode], Set[ChoiceNode]]:

    if traversed is None:
        traversed = set(start_nodes)

    # Get next step nodes
    next_nodes = set()
    for node in start_nodes:
        next_nodes |= {edge[1] for edge in iter_out_edges(graph, node)
                       if get_edge_type(edge) in [EdgeType.DERIVES, EdgeType.CONNECTS]}
    next_nodes -= traversed
    traversed |= next_nodes

    # Get choice nodes
    choice_nodes = set()
    non_decision_nodes = set()
    for node in next_nodes:
        if isinstance(node, ChoiceNode):
            choice_nodes.add(node)
        else:
            non_decision_nodes.add(node)

    # Recursively get choice nodes
    if len(non_decision_nodes) > 0:
        next_non_decision_nodes, next_decision_nodes = \
            traverse_until_choice_nodes(graph, non_decision_nodes, traversed=traversed)
        non_decision_nodes |= next_non_decision_nodes
        choice_nodes |= next_decision_nodes

    return (non_decision_nodes | start_nodes), choice_nodes


def get_non_confirmed_nodes(graph: nx.MultiDiGraph, start_nodes: Set[DSGNode]) -> Set[DSGNode]:

    # Get confirmed nodes
    confirmed_nodes, _ = traverse_until_choice_nodes(graph, set(start_nodes))

    # Get non-confirmed nodes
    non_confirmed_nodes = set(graph.nodes) - confirmed_nodes
    return non_confirmed_nodes


def iter_in_edges_cached(graph: nx.MultiDiGraph, node: 'DSGNode', edge_type: EdgeType = None,
                         cache=None) -> Iterator['EdgeTuple']:

    if cache is None:
        yield from iter_in_edges(graph, node, edge_type=edge_type)
        return

    cache['iter_in'] = iter_in_cache = cache.get('iter_in', {})

    in_edges = iter_in_cache.get(node)
    if in_edges is not None:
        yield from in_edges
        return

    iter_in_cache[node] = in_edges = list(iter_in_edges(graph, node))
    yield from in_edges


def iter_out_edges_cached(graph: nx.MultiDiGraph, node: 'DSGNode', edge_type: EdgeType = None,
                          cache=None) -> Iterator['EdgeTuple']:

    if cache is None:
        yield from iter_out_edges(graph, node, edge_type=edge_type)
        return

    cache['iter_out'] = iter_out_cache = cache.get('iter_out', {})

    out_edges = iter_out_cache.get(node)
    if out_edges is not None:
        yield from out_edges
        return

    iter_out_cache[node] = out_edges = list(iter_out_edges(graph, node))
    yield from out_edges



"""
MIT License

Copyright: (c) 2024, Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
Contact: jasper.bussemaker@dlr.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import os
import uuid
import json
import tempfile
import webbrowser
import numpy as np
from typing import *
from enum import Enum, auto
from adsg_core.graph.export import *
from adsg_core.graph.adsg_basic import *
from adsg_core.graph.adsg_nodes import *
from adsg_core.graph.graph_edges import *
from adsg_core.graph.choice_constraints import *
from adsg_core.optimization.graph_processor import GraphProcessor

__all__ = ['DSGRenderer', 'Legend']


class Legend(Enum):
    NODES = auto()
    EDGES = auto()
    CHOICES = auto()
    METRICS = auto()

    NODE = auto()
    NODE_START = auto()
    NODE_CONNECTOR = auto()

    EDGE_DERIVE = auto()
    EDGE_CONNECT = auto()
    EDGE_INCOMP = auto()
    EDGE_EXCLUDE = auto()

    CHOICE_SEL = auto()
    CHOICE_CONN = auto()

    METRICS_DV = auto()
    METRICS_OUTPUTS = auto()

    CHOICE_CONSTRAINT = auto()


class DSGRenderer:
    """Utility class for rendering and displaying DSGs"""

    _local_viz_js = False

    def __init__(self, dsg: DSGType, title=None):
        self._dsg = dsg
        self._title = title if title is not None else 'DSG'

    def render(self, path=None, title=None, print_svg=False, print_dot=False):
        """
        Render the DSG and display it in a Jupyter notebook.
        """
        self._render(self._dsg, title=title or self._title, path=path, print_svg=print_svg, print_dot=print_dot)

    def render_all_instances(self, idx=None, title=None, print_svg=False, print_dot=False):
        from IPython.display import display, Markdown

        processor = GraphProcessor(self._dsg)
        x_all, _ = processor.get_all_discrete_x()

        n_total = x_all.shape[0]
        idx_all = np.arange(n_total)
        if idx is not None:
            x_all = x_all[idx, :]
            idx_all = idx_all[idx]
            display(Markdown(f'Rendering {x_all.shape[0]} of {n_total} instances'))
        else:
            display(Markdown(f'Rendering {n_total} instances'))

        if title is None:
            title = self._title
        for i, xi in enumerate(x_all):
            graph, _, _ = processor.get_graph(xi)
            self._render(graph, title=f'{title} [{idx_all[i]+1}/{x_all.shape[0]}]',
                         print_svg=print_svg, print_dot=print_dot, repeated=i > 0)

    @classmethod
    def _render(cls, dsg: DSGType, title, path=None, print_svg=False, print_dot=False, repeated=False):
        """
        Render the DSG and display it in a Jupyter notebook.
        """
        dot_contents = dsg.export_dot()
        cls._render_dot(dot_contents, title, path=path, print_svg=print_svg, print_dot=print_dot, repeated=repeated)

    @classmethod
    def _render_dot(cls, dot_contents, title, path=None, print_svg=False, print_dot=False, repeated=False):
        dot_html = cls._render_html(dot_contents, print_svg=print_svg, print_dot=print_dot, repeated=repeated)
        if cls._running_in_ipython():
            cls._display_ipython(dot_html)
        else:
            cls._display_browser(dot_html, title, path=path)

    @classmethod
    def _viz_js_tag(cls):
        if cls._local_viz_js:
            path = os.path.join(os.path.dirname(__file__), 'resources', 'viz-standalone.js')
            if os.path.exists(path):
                # return f'file:///{path}'
                with open(path, 'r') as fp:
                    js_contents = fp.read()
                    return (f'<script type="text/javascript">'
                            f'{ js_contents }'
                            f'</script>')

        return ('<script type="text/javascript" src="'
                'https://cdn.jsdelivr.net/npm/@viz-js/viz@3.8.0/lib/viz-standalone.js"></script>')

    @classmethod
    def _render_html(cls, dot, print_svg=False, print_dot=False, repeated=False):
        div_id = uuid.uuid4().hex
        return f"""<div id="{div_id}"></div>
{cls._viz_js_tag() if not repeated else ""}
<script type="text/javascript">
(function() {{
  var dot = {json.dumps(dot)}; // Export of the dot graph notation
  var printDot = {json.dumps(print_dot)}; // Whether to print the dot instead of rendering
  var printSvg = {json.dumps(print_svg)}; // Whether to print the SVG instead of showing the results
  // Create viz-js instance and render to SVG
  function doRender() {{
    if (printDot) {{
      var preEl = document.createElement('pre');
      preEl.innerText = dot;
      document.getElementById("{div_id}").appendChild(preEl);
    }} else {{
      Viz.instance().then(function(viz) {{
        var tgtEl = document.getElementById("{div_id}");
        var svgElement = viz.renderSVGElement(dot);
        if (printSvg) {{
          var preEl = document.createElement('pre');
          preEl.innerText = svgElement.outerHTML;
          tgtEl.appendChild(preEl);
        }} else {{
          tgtEl.appendChild(svgElement);
        }}
      }}).catch(function(e) {{
        console.log('RENDER ERROR', e, dot);
        var preEl = document.createElement('pre');
        preEl.innerText = 'RENDER ERROR: '+e.toString();
        document.getElementById("{div_id}").appendChild(preEl);
        var preEl = document.createElement('pre');
        preEl.innerText = dot;
        document.getElementById("{div_id}").appendChild(preEl);
      }});
    }}
  }}
  // We may need to wait for loading to complete
  function checkRender() {{
    if (typeof Viz === "undefined") {{ setTimeout(checkRender, 100+200*Math.random()); }} else {{ doRender(); }} 
  }}
  checkRender();
}})()
</script>
"""

    @staticmethod
    def _running_in_ipython():
        try:
            from IPython.core.interactiveshell import InteractiveShell
            return InteractiveShell.initialized()
        except ModuleNotFoundError:
            return False

    @staticmethod
    def _display_ipython(dot_html):
        from IPython.display import display, HTML
        display(HTML(dot_html))

    @staticmethod
    def _display_browser(dot_html, title, path=None):
        full_html = f"""<!doctype html>
<html><head><title>{title}</title></head>
<body>{dot_html}</body></html>"""

        if path is None:
            with tempfile.NamedTemporaryFile('w', delete=False, suffix='.html') as fp:
                fp.write(full_html)
                url = f'file://{fp.name}'
        else:
            with open(path, 'w') as fp:
                fp.write(full_html)
                url = f'file://{path}'

        webbrowser.open(url)

    @classmethod
    def render_legend(cls, path=None, elements: Set[Union[str, Legend]] = None, print_svg=False, print_dot=False):
        dot_contents = cls._render_legend_dot(elements=elements).to_string()
        cls._render_dot(dot_contents, 'DSG Legend', path=path, print_svg=print_svg, print_dot=print_dot)

    @classmethod
    def _render_legend_dot(cls, elements: Set[Legend] = None):
        dsg = cls._get_legend_dsg(elements=elements)
        dot_graph = export_dot(dsg.graph, start_nodes=dsg.derivation_start_nodes,
                               choice_constraints=dsg.get_choice_constraints(), legend_mode=True, return_dot=True)

        dot_graph.obj_dict['attributes'].update(dict(
            label='DSG Legend',
            labelloc='t',
            fontsize='16',
            bgcolor='#f5f8fa',
        ))

        return dot_graph

    @staticmethod
    def _get_legend_dsg(elements: Set[Union[str, Legend]] = None):
        if elements is None:
            elements = {Legend.NODES, Legend.EDGES, Legend.CHOICES, Legend.METRICS, Legend.CHOICE_CONSTRAINT}
        else:
            elements = {el if isinstance(el, Legend) else Legend[el.upper()] for el in elements}
        if Legend.NODES in elements:
            elements |= {Legend.NODE, Legend.NODE_CONNECTOR, Legend.NODE_START}
        if Legend.EDGES in elements:
            elements |= {Legend.EDGE_DERIVE, Legend.EDGE_CONNECT, Legend.EDGE_INCOMP, Legend.EDGE_EXCLUDE}
        if Legend.CHOICES in elements:
            elements |= {Legend.CHOICE_SEL, Legend.CHOICE_CONN}
        if Legend.METRICS in elements:
            elements |= {Legend.METRICS_DV, Legend.METRICS_OUTPUTS}

        dsg = BasicDSG()
        dsg._start_nodes = set()

        if Legend.CHOICE_CONSTRAINT in elements:
            c1, c2 = SelectionChoiceNode('Choice 1'), SelectionChoiceNode('Choice 2')
            dsg.add_node(c1)
            dsg.add_node(c2)
            dsg.constrain_choices(ChoiceConstraintType.LINKED, [c1, c2])

        if Legend.NODE_CONNECTOR in elements:
            dsg.add_node(ConnectorNode('Connector'))
        if Legend.NODE in elements:
            dsg.add_node(NamedNode('Node'))

        # if Legend.CHOICE_CONSTRAINT in elements:
        #     for name, constraint in [
        #         ('Unordered\\nNon-replacing\\nConstraint', ChoiceConstraintType.UNORDERED_NOREPL),
        #         ('Unordered\\nConstraint', ChoiceConstraintType.UNORDERED),
        #         ('Permutation\\nConstraint', ChoiceConstraintType.PERMUTATION),
        #         ('Linked\\nConstraint', ChoiceConstraintType.LINKED),
        #     ]:
        #         src, opts = NamedNode(name), [NamedNode('Opt 1'), NamedNode('Opt 2')]
        #         dsg.constrain_choices(constraint, [
        #             dsg.add_selection_choice('Choice 1', src, opts),
        #             dsg.add_selection_choice('Choice 2', src, opts),
        #         ])

        if len(elements & {Legend.NODE_START, Legend.METRICS_DV, Legend.METRICS_OUTPUTS}) > 0:
            start_node = NamedNode('Start Node')
            dsg.add_node(start_node)
            dsg._start_nodes = {start_node}
            if Legend.METRICS_DV in elements:
                dsg.add_edge(start_node, DesignVariableNode('Continuous\\nDV', bounds=(0, 1)))
                dsg.add_edge(start_node, DesignVariableNode('Discrete\\nDV', options=['A', 'B', 'C']))
            if Legend.METRICS_OUTPUTS in elements:
                dsg.add_edge(start_node, MetricNode('Objective', direction=-1))
                dsg.add_edge(start_node, MetricNode('Constraint', direction=-1, ref=0))
                dsg.add_edge(start_node, MetricNode('Output Metric'))

        if Legend.CHOICE_CONN in elements:
            src_nodes = [
                ConnectorNode('Source 1', deg_list=[1]),
                ConnectorNode('Source 2', deg_list=[0, 1]),
                ConnectorNode('Source 2', deg_min=0, repeated_allowed=True),
            ]
            tgt_nodes = [
                ConnectorNode('Target 1', deg_min=1, repeated_allowed=True),
                ConnectorNode('Target 2', deg_min=0, deg_max=2),
            ]
            dsg.add_connection_choice(
                'Connection\\nChoice',
                src_nodes=src_nodes,
                tgt_nodes=tgt_nodes,
                exclude=[(src_nodes[0], tgt_nodes[0])] if Legend.EDGE_EXCLUDE in elements else [],
            )

        if Legend.CHOICE_SEL in elements:
            dsg.add_selection_choice('Selection\\nChoice', NamedNode('Source'),
                                      [NamedNode('Option 1'), NamedNode('Option 2'), NamedNode('Etc...')])

        if Legend.EDGE_INCOMP in elements:
            add_edge(dsg.graph, NamedNode('E'), NamedNode('F'), edge_type=EdgeType.INCOMPATIBILITY,
                     label='mutually\\nincompatible')
        if Legend.EDGE_CONNECT in elements:
            add_edge(dsg.graph, NamedNode('C'), NamedNode('D'), edge_type=EdgeType.CONNECTS, label='connects to')
        if Legend.EDGE_DERIVE in elements:
            add_edge(dsg.graph, NamedNode('A'), NamedNode('B'), edge_type=EdgeType.DERIVES, label='derives')

        return dsg



import os
import math
import numba
import pickle
import hashlib
import itertools
import functools
import numpy as np
from typing import *
from dataclasses import dataclass
from adsg_core.optimization.assign_enc.cache import get_cache_path

__all__ = ['Node', 'NodeExistence', 'NodeExistencePatterns', 'AggregateAssignmentMatrixGenerator', 'count_n_pool_take',
           'MatrixMap', 'MatrixMapOptional', 'count_src_to_target', 'MatrixGenSettings']


class Node:
    """
    Defines a source or target node with a specified number of possible outgoing/incoming connections:
    either an explicit list or only a lower bound.
    """

    def __init__(self, nr_conn_list: List[int] = None, min_conn: int = None, max_conn: int = math.inf,
                 repeated_allowed: bool = True):

        # Flag whether repeated connections to the same opposite node is allowed
        self.rep = repeated_allowed

        # Determine whether we have a set list of connections or only a lower bound
        conns, min_conns = None, None
        if nr_conn_list is not None:
            conns = nr_conn_list

        elif min_conn is not None:
            if max_conn == math.inf:
                min_conns = min_conn
            else:
                conns = list(range(min_conn, max_conn+1))
        else:
            raise ValueError('Either supply a list or a lower limit')
        self.conns = sorted(conns) if conns is not None else None
        self.min_conns = min_conns

    @property
    def max_inf(self):
        return self.conns is None

    def __str__(self):
        if self.conns is None:
            return f'({self.min_conns}..*)'
        return f'({",".join([str(i) for i in self.conns])})'

    def __repr__(self):
        return f'{self.__class__.__name__}(conns={self.conns!r}, min_conns={self.min_conns}, rep={self.rep})'


class NodeExistence:
    """
    Defines a specific src/tgt node existence pattern.
    """

    def __init__(
            self,
            src_exists: Union[List[bool], np.ndarray] = None,
            tgt_exists: Union[List[bool], np.ndarray] = None,
            src_n_conn_override: Dict[int, List[int]] = None,
            tgt_n_conn_override: Dict[int, List[int]] = None,
            max_src_conn_override: int = None,
            max_tgt_conn_override: int = None,
    ):
        self.src_n_conn_override = self._get_n_conn_override(src_exists, src_n_conn_override)
        self.tgt_n_conn_override = self._get_n_conn_override(tgt_exists, tgt_n_conn_override)
        self.max_src_conn_override = max_src_conn_override
        self.max_tgt_conn_override = max_tgt_conn_override
        self._hash = None

        self._src_exists_mask = None
        self._tgt_exists_mask = None
        self._none_exists = None

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_hash'] = None
        return state

    def __setstate__(self, state):
        state['_hash'] = None
        self.__dict__ = state

    @staticmethod
    def _get_n_conn_override(exists: Union[List[bool], np.ndarray] = None,
                             n_conn_override: Dict[int, List[int]] = None) -> Dict[int, List[int]]:
        if n_conn_override is None and exists is None:
            return {}

        n_conn_override = {} if n_conn_override is None else n_conn_override.copy()
        if exists is not None:
            for i, exist in enumerate(exists):
                if not exist:
                    n_conn_override[i] = [0]
        if len(n_conn_override) == 0:
            return {}
        return n_conn_override

    def has_src(self, i):
        return self.src_n_conn_override.get(i) != [0]

    def has_tgt(self, i):
        return self.tgt_n_conn_override.get(i) != [0]

    def src_exists_mask(self, n_src: int):
        if self._src_exists_mask is not None:
            return self._src_exists_mask

        exist_mask = np.ones((n_src,), dtype=bool)
        for i in range(n_src):
            if not self.has_src(i):
                exist_mask[i] = False
        self._src_exists_mask = exist_mask
        return exist_mask

    def tgt_exists_mask(self, n_tgt: int):
        if self._tgt_exists_mask is not None:
            return self._tgt_exists_mask

        exist_mask = np.ones((n_tgt,), dtype=bool)
        for i in range(n_tgt):
            if not self.has_tgt(i):
                exist_mask[i] = False
        self._tgt_exists_mask = exist_mask
        return exist_mask

    def none_exists(self, n_src: int, n_tgt: int):
        def _none_exists():
            if np.any(self.src_exists_mask(n_src)):
                return False
            if np.any(self.tgt_exists_mask(n_tgt)):
                return False
            return True

        if self._none_exists is None:
            self._none_exists = _none_exists()
        return self._none_exists

    def get_effective_settings(self, settings: 'MatrixGenSettings') \
            -> Tuple['MatrixGenSettings', Dict[int, int], Dict[int, int]]:
        """Returns connection settings with all connection degree overrides of this existence pattern applied"""

        def _get_effective_nodes(nodes: List[Node], n_conn_override: Dict[int, List[int]] = None,
                                 max_conn_override: int = None) -> Tuple[List[Node], Dict[int, int]]:
            effective_nodes = []
            idx_map = {}
            for i, node in enumerate(nodes):
                # Get effective number of connections
                n_conns, min_conn, max_conn = None, None, math.inf
                if i in n_conn_override:
                    n_conns = n_conn_override[i]

                elif max_conn_override is not None:
                    if node.max_inf:
                        min_conn, max_conn = node.min_conns, max_conn_override
                    else:
                        n_conns = [n for n in node.conns if n <= max_conn_override]

                else:
                    n_conns, min_conn = node.conns, node.min_conns

                # Create effective node and check if it is not inactive
                effective_node = Node(n_conns, min_conn=min_conn, max_conn=max_conn, repeated_allowed=node.rep)
                if not effective_node.max_inf and (len(effective_node.conns) == 0 or effective_node.conns == [0]):
                    continue

                idx_map[i] = len(effective_nodes)
                effective_nodes.append(effective_node)

            return effective_nodes, idx_map

        src, src_idx_map = _get_effective_nodes(settings.src, self.src_n_conn_override, self.max_src_conn_override)
        tgt, tgt_idx_map = _get_effective_nodes(settings.tgt, self.tgt_n_conn_override, self.max_tgt_conn_override)

        # Automatically convert to bounded connection nrs if needed
        # For this we need to fix the maximum nr of connections before the conversion
        max_conn_par = MatrixGenSettings(
            src=src, tgt=tgt, max_conn_parallel=settings.max_conn_parallel).get_max_conn_parallel()

        def _get_max_outgoing_conn(ref_nodes: List[Node]):
            n_max_outgoing = 0
            for n in ref_nodes:
                if not n.rep:
                    n_max_outgoing += 1
                elif n.conns is not None:
                    n_max_outgoing += min(max_conn_par, n.conns[-1])
                else:
                    return -1  # If there is a node with no upper bound, we cannot reliably check the max nr of conns
            return n_max_outgoing

        for nodes, n_conn_max in [(tgt, _get_max_outgoing_conn(src)),
                                  (src, _get_max_outgoing_conn(tgt))]:
            if n_conn_max < 1:
                continue
            n_conn_max = max(2, n_conn_max)
            for node in nodes:
                if node.conns is not None:
                    for min_conns in range(n_conn_max):
                        consecutive_n_conns = list(range(min_conns, n_conn_max+1))
                        if node.conns[:len(consecutive_n_conns)] == consecutive_n_conns:
                            # print(f'CONVERTING {node.conns} -> min {min_conns}; {node!r}')
                            node.min_conns = min_conns
                            node.conns = None
                            break

        # Modify excluded edges
        excluded = []
        for i_src, i_tgt in settings.get_excluded_indices():
            # If source or target node are disabled, skip
            if i_src not in src_idx_map or i_tgt not in tgt_idx_map:
                continue

            excluded.append((src_idx_map[i_src], tgt_idx_map[i_tgt]))

        effective_settings = MatrixGenSettings(
            src=src, tgt=tgt, excluded=excluded, max_conn_parallel=max_conn_par)
        return effective_settings, src_idx_map, tgt_idx_map

    def get_transpose(self) -> 'NodeExistence':
        return self.__class__(
            src_n_conn_override=self.tgt_n_conn_override, tgt_n_conn_override=self.src_n_conn_override,
            max_src_conn_override=self.max_tgt_conn_override, max_tgt_conn_override=self.max_src_conn_override)

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((
                tuple([(k, tuple(v)) for k, v in self.src_n_conn_override.items()]) if self.src_n_conn_override is not None else -1,
                tuple([(k, tuple(v)) for k, v in self.tgt_n_conn_override.items()]) if self.tgt_n_conn_override is not None else -1,
                self.max_src_conn_override, self.max_tgt_conn_override,
            ))
        return self._hash

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __repr__(self):
        src_conn_str = '' if self.src_n_conn_override is None else ';'.join([f'{k}:{",".join([str(vv) for vv in v])}' for k, v in self.src_n_conn_override.items()])
        tgt_conn_str = '' if self.tgt_n_conn_override is None else ';'.join([f'{k}:{",".join([str(vv) for vv in v])}' for k, v in self.tgt_n_conn_override.items()])
        max_src_conn_str = f' ms{self.max_src_conn_override}' if self.max_src_conn_override is not None else ''
        max_tgt_conn_str = f' mt{self.max_tgt_conn_override}' if self.max_tgt_conn_override is not None else ''
        return f'{src_conn_str} _ {tgt_conn_str}{max_src_conn_str}{max_tgt_conn_str}'


@dataclass(frozen=False)
class NodeExistencePatterns:
    patterns: List[NodeExistence]

    def __post_init__(self):
        if len(set(self.patterns)) != len(self.patterns):
            raise RuntimeError('Duplicate node existence patterns')

    @classmethod
    def always_exists(cls) -> 'NodeExistencePatterns':
        return cls([NodeExistence()])

    @classmethod
    def get_all_combinations(cls, src_is_conditional: List[bool], tgt_is_conditional: List[bool]):
        """
        Gets patterns for all combination of conditional existence, e.g.:
        is_conditional = [False, True, True]
        exists = [True, False, False]
        exists = [True, True , False]
        exists = [True, False, True ]
        exists = [True, True , True ]
        """
        src_exist_opts = [[True, False] if is_conditional else [True] for is_conditional in src_is_conditional]
        tgt_exist_opts = [[True, False] if is_conditional else [True] for is_conditional in tgt_is_conditional]

        patterns = []
        for src_exists in itertools.product(*src_exist_opts):
            for tgt_exists in itertools.product(*tgt_exist_opts):
                patterns.append(NodeExistence(src_exists=list(src_exists), tgt_exists=list(tgt_exists)))
        return cls(patterns)

    @classmethod
    def get_increasing(cls, src_is_conditional: List[bool], tgt_is_conditional: List[bool]):
        """
        Gets patterns for conditional existence combinations where node existence is monotonically increasing, e.g.:
        is_conditional = [False, True, True]
        exists = [True, False, False]
        exists = [True, True , False]
        exists = [True, True , True ]
        """

        def _get_exist_opts(is_conditional: List[bool]) -> List[List[bool]]:
            n_cond = sum([is_cond for is_cond in is_conditional])
            exist_opts = []
            for n_cond_exists in range(n_cond+1):
                n_exists = 0
                exist_opt = []
                for is_cond in is_conditional:
                    if is_cond:
                        exist_opt.append(n_exists < n_cond_exists)
                        n_exists += 1
                    else:
                        exist_opt.append(True)
                exist_opts.append(exist_opt)
            return exist_opts

        src_exist_opts = _get_exist_opts(src_is_conditional)
        tgt_exist_opts = _get_exist_opts(tgt_is_conditional)
        return cls([NodeExistence(src_exists=list(src_exists), tgt_exists=list(tgt_exists))
                    for src_exists, tgt_exists in itertools.product(src_exist_opts, tgt_exist_opts)])

    def src_is_conditional(self, i):
        for pattern in self.patterns:
            if not pattern.has_src(i):
                return True
        return False

    def tgt_is_conditional(self, i):
        for pattern in self.patterns:
            if not pattern.has_tgt(i):
                return True
        return False


@dataclass
class MatrixGenSettings:
    src: List[Node]
    tgt: List[Node]
    excluded: List[Union[Tuple[Node, Node], Tuple[int, int]]] = None
    existence: NodeExistencePatterns = None
    max_conn_parallel: int = None

    def get_max_conn_parallel(self):
        """If not given, calculate it by taking the nr such that each target or source can receive at least the minimum
        nr of connections, but at least 2 per connection"""
        if self.max_conn_parallel is not None:
            return max(1, self.max_conn_parallel)

        max_non_inf = 2
        for nodes in [self.src, self.tgt]:
            for node in nodes:
                if not node.max_inf:
                    max_conn = max(node.conns)
                    if max_non_inf is None or max_conn > max_non_inf:
                        max_non_inf = max_conn

        return max_non_inf

    def get_excluded_indices(self) -> List[Tuple[int, int]]:
        excluded_edges = []
        src_map = {node: i for i, node in enumerate(self.src)}
        tgt_map = {node: i for i, node in enumerate(self.tgt)}
        for src, tgt in (self.excluded or []):
            excluded_edges.append((src_map.get(src, src), tgt_map.get(tgt, tgt)))
        return excluded_edges

    def get_max_conn_matrix(self):
        # Restrict by maximum parallel connections
        max_conns = np.ones((len(self.src), len(self.tgt)), dtype=int)*self.get_max_conn_parallel()

        # Restrict by max fixed connection numbers
        for i_src, node in enumerate(self.src):
            if not node.max_inf:
                max_conns[i_src, :] = np.min([max_conns[i_src, :], [max(node.conns)]*len(self.tgt)], axis=0)
        for i_tgt, node in enumerate(self.tgt):
            if not node.max_inf:
                max_conns[:, i_tgt] = np.min([max_conns[:, i_tgt], [max(node.conns)]*len(self.src)], axis=0)

        # Restrict by non-repeatability into/out of nodes
        for i_src, node in enumerate(self.src):
            if not node.rep:
                max_conns[i_src, max_conns[i_src, :] > 1] = 1
        for i_tgt, node in enumerate(self.tgt):
            if not node.rep:
                max_conns[max_conns[:, i_tgt] > 1, i_tgt] = 1

        # Restrict excluded edges
        for i_src, i_tgt in self.get_excluded_indices():
            max_conns[i_src, i_tgt] = 0

        return max_conns

    def get_effective_settings(self) -> Dict[NodeExistence, Tuple['MatrixGenSettings', Dict[int, int], Dict[int, int]]]:
        """For each node existence pattern, get the effective node connection settings"""
        patterns = self.existence.patterns if self.existence is not None else [NodeExistence()]
        return {existence: existence.get_effective_settings(self) for existence in patterns}

    def expand_effective_matrix(self, matrix: np.ndarray, src_map: Dict[int, int], tgt_map: Dict[int, int]) -> np.ndarray:
        if len(src_map) == len(self.src) and len(tgt_map) == len(self.tgt):
            return matrix

        is_2d = len(matrix.shape) == 2
        if is_2d:
            matrix = np.array([matrix])
        expanded_matrix = np.zeros((matrix.shape[0], len(self.src), len(self.tgt)), dtype=int)
        tgt_indices = list(tgt_map.keys())
        for i_src in range(len(self.src)):
            if i_src in src_map:
                expanded_matrix[:, i_src, tgt_indices] = matrix[:, src_map[i_src], :]

        return expanded_matrix[0, :, :] if is_2d else expanded_matrix

    def get_transpose_settings(self) -> 'MatrixGenSettings':
        tgt, src = self.src, self.tgt
        excluded = [(j, i) for i, j in self.get_excluded_indices()] if self.excluded is not None else None

        existence_patterns = None
        if self.existence is not None:
            existence_patterns = NodeExistencePatterns(patterns=[
                existence.get_transpose() for existence in self.existence.patterns])

        return MatrixGenSettings(src=src, tgt=tgt, excluded=excluded, existence=existence_patterns,
                                 max_conn_parallel=self.max_conn_parallel)

    def get_cache_key(self):
        cache_version = '2'  # Increment this if all existing caches should be invalidated

        src_cache_key = ';'.join([repr(s) for s in self.src])
        tgt_cache_key = ';'.join([repr(t) for t in self.tgt])
        excluded_cache_key = ';'.join([f'{tup[0]:d},{tup[1]:d}' for tup in sorted([ex for ex in self.get_excluded_indices()])]) \
            if self.excluded is not None else ''
        exist_cache_key = ';'.join([str(hash(p)) for p in self.existence.patterns]) if self.existence is not None else ''
        cache_str = '||'.join([src_cache_key, tgt_cache_key, excluded_cache_key, exist_cache_key,
                               str(self.max_conn_parallel), cache_version])
        return hashlib.md5(cache_str.encode('utf-8')).hexdigest()


MatrixMap = Dict[NodeExistence, np.ndarray]
MatrixMapOptional = Union[np.ndarray, Dict[NodeExistence, np.ndarray]]


class AggregateAssignmentMatrixGenerator:
    """
    Generator that iterates over all connection possibilities between two sets of objects, with various constraints:
    - Specified number or range of connections per object
    - Repeated connections between objects allowed or not
    - List of excluded connections
    """

    def __init__(self, settings: MatrixGenSettings):
        self._settings = settings
        self._exist = None
        self._max_conn_mat = None
        self._max_src_appear = {}
        self._max_tgt_appear = {}
        self._node_settings = None
        self._n_conn_override_map = {}

    def __setstate__(self, state):
        if '_max_conn_mat' in state and isinstance(state['_max_conn_mat'], np.ndarray):
            state['_max_conn_mat'] = None
        self.__dict__.update(**state)

    @classmethod
    def create(cls, src: List[Node], tgt: List[Node], **kwargs):
        return cls(MatrixGenSettings(src, tgt, **kwargs))

    @property
    def settings(self) -> MatrixGenSettings:
        return self._settings

    @property
    def src(self):
        return self._settings.src

    @property
    def tgt(self):
        return self._settings.tgt

    @property
    def ex(self) -> Optional[List[Tuple[int, int]]]:
        if self._settings.existence is not None:
            return self._settings.get_excluded_indices()

    @property
    def existence_patterns(self) -> NodeExistencePatterns:
        if self._exist is None:
            self._exist = self._settings.existence or NodeExistencePatterns.always_exists()
        return self._exist

    @property
    def max_conn_mat(self) -> Dict[NodeExistence, np.ndarray]:
        if self._max_conn_mat is None:
            self._max_conn_mat = max_conn_mat_map = {}
            for existence, (settings, src_map, tgt_map) in self._settings.get_effective_settings().items():
                max_conn_mat_map[existence] = \
                    self._settings.expand_effective_matrix(settings.get_max_conn_matrix(), src_map, tgt_map)
        return self._max_conn_mat

    def get_max_conn_mat(self, existence: NodeExistence) -> np.ndarray:
        if existence not in self.max_conn_mat:
            return self.max_conn_mat[NodeExistence()]
        return self.max_conn_mat[existence]

    def get_max_src_appear(self, existence: NodeExistence = None):
        if existence is None:
            existence = NodeExistence()
        if existence not in self._max_src_appear:
            max_appear = np.sum(self.get_max_conn_mat(existence), axis=1)
            if existence.max_src_conn_override is not None:
                max_appear[max_appear > existence.max_src_conn_override] = existence.max_src_conn_override
            self._max_src_appear[existence] = max_appear
        return self._max_src_appear[existence]

    def get_max_tgt_appear(self, existence: NodeExistence = None):
        if existence is None:
            existence = NodeExistence()
        if existence not in self._max_tgt_appear:
            max_appear = np.sum(self.get_max_conn_mat(existence), axis=0)
            if existence.max_tgt_conn_override is not None:
                max_appear[max_appear > existence.max_tgt_conn_override] = existence.max_tgt_conn_override
            self._max_tgt_appear[existence] = max_appear
        return self._max_tgt_appear[existence]

    def __iter__(self) -> Generator[Tuple[np.array, NodeExistence], None, None]:
        for n_src_conn, n_tgt_conn, existence in self.iter_n_sources_targets():
            yield self._iter_matrices(n_src_conn, n_tgt_conn, existence), existence  # Iterate over assignment matrices

    def iter_matrices(self, existence: NodeExistence = None):
        for n_src_conn, n_tgt_conn, exist in self.iter_n_sources_targets(existence=existence):
            for matrix in self._iter_matrices(n_src_conn, n_tgt_conn, existence=exist):
                yield matrix, exist

    def get_agg_matrix(self, cache=False) -> MatrixMap:
        if cache:
            agg_matrix = self._load_agg_matrix_from_cache()
            if agg_matrix is not None:
                return agg_matrix

        agg_matrix = self._agg_matrices(self, ensure_all_existence=True)
        self._write_to_cache(self._get_cache_file(), agg_matrix)
        return agg_matrix

    def _load_agg_matrix_from_cache(self) -> Optional[MatrixMap]:
        return self._load_from_cache(self._get_cache_file())

    @staticmethod
    def _write_to_cache(cache_path, obj):
        os.makedirs(os.path.dirname(cache_path), exist_ok=True)
        with open(cache_path, 'wb') as fp:
            pickle.dump(obj, fp)

    @staticmethod
    def _load_from_cache(cache_path):
        if os.path.exists(cache_path):
            with open(cache_path, 'rb') as fp:
                return pickle.load(fp)

    def reset_agg_matrix_cache(self):
        cache_path = self._get_cache_file()
        if os.path.exists(cache_path):
            os.remove(cache_path)

        cache_path = self._get_cache_file_iter_n()
        if os.path.exists(cache_path):
            os.remove(cache_path)

    def _get_cache_file(self):
        return self._cache_path(f'{self._get_cache_key()}.pkl')

    def _get_cache_file_iter_n(self):
        return self._cache_path(f'{self._get_cache_key()}_iter_n.pkl')

    def _get_cache_key(self):
        return self.settings.get_cache_key()

    @staticmethod
    def _cache_path(sub_path=None):
        sel_cache_folder = 'matrix_cache'
        sub_path = os.path.join(sel_cache_folder, sub_path) if sub_path is not None else sel_cache_folder
        return get_cache_path(sub_path=sub_path)

    def _agg_matrices(self, matrix_gen, ensure_all_existence=False):
        """Aggregate generated matrices into one big matrix, one per existence pattern"""
        matrices_by_existence = {}
        for matrix_data in list(matrix_gen):
            if isinstance(matrix_data, np.ndarray):
                matrices, existence = matrix_data, NodeExistence()
            else:
                matrices, existence = matrix_data

            if existence not in matrices_by_existence:
                matrices_by_existence[existence] = []
            matrices_by_existence[existence].append(matrices)

        if ensure_all_existence:
            for existence in self.iter_existence():
                if existence not in matrices_by_existence:
                    matrices_by_existence[existence] = []

        agg_matrices_by_existence = {}
        for existence, matrices in matrices_by_existence.items():
            if len(matrices) == 0:
                agg_matrices_by_existence[existence] = np.zeros((0, len(self.src), len(self.tgt)), dtype=int)
            elif len(matrices) == 1:
                agg_matrices_by_existence[existence] = matrices[0]
            else:
                agg_matrices_by_existence[existence] = np.row_stack(matrices)
        return agg_matrices_by_existence

    def get_matrix_for_existence(self, matrix: MatrixMapOptional, existence: NodeExistence = None) -> np.ndarray:
        if isinstance(matrix, np.ndarray):
            return matrix
        if existence not in matrix:
            return np.empty((0, len(self.src), len(self.tgt)), dtype=int)
        return matrix[existence]

    def iter_conns(self) -> Generator[List[Tuple[Node, Node]], None, None]:
        """Generate lists of edges from matrices"""
        for matrix, _ in self.iter_matrices():
            yield tuple(self.get_conns(matrix))

    @staticmethod
    def get_conn_idx(matrix: np.ndarray) -> List[Tuple[int, int]]:
        """Convert matrix to edge tuples"""
        edges = []
        for i_src in range(matrix.shape[0]):
            for j_tgt in range(matrix.shape[1]):
                for _ in range(matrix[i_src, j_tgt]):
                    edges.append((i_src, j_tgt))
        return edges

    def get_conns(self, matrix: np.ndarray) -> List[Tuple[Node, Node]]:
        """Convert matrix to edge tuples"""
        edges = []
        for i_src in range(matrix.shape[0]):
            src = self.src[i_src]
            for j_tgt in range(matrix.shape[1]):
                tgt = self.tgt[j_tgt]
                for _ in range(matrix[i_src, j_tgt]):
                    edges.append((src, tgt))
        return edges

    def iter_n_sources_targets(self, existence: NodeExistence = None, cache=True):
        cache_path = self._get_cache_file_iter_n()
        if cache:
            iter_data = self._load_from_cache(cache_path)
            if iter_data is not None:
                n_src = len(self.src)
                for exist, np_tup in iter_data:
                    if existence is None or existence == exist:
                        for i in range(np_tup.shape[0]):
                            yield tuple(np_tup[i, :n_src]), tuple(np_tup[i, n_src:]), exist
                return

        tuples = {}
        for n_src_conn, n_tgt_conn, exist in self._iter_n_sources_targets():
            if exist not in tuples:
                tuples[exist] = []
            tuples[exist].append((n_src_conn, n_tgt_conn))
            if existence is None or existence == exist:
                yield n_src_conn, n_tgt_conn, exist

        # Store it as a numpy array (more efficient than storing as tuples)
        tuples_efficient = []
        n_src = len(self.src)
        for exist, n_src_n_tgt in tuples.items():
            np_tup = np.empty((len(n_src_n_tgt), n_src+len(self.tgt)), dtype=np.int64)
            for i, (n_src_conn, n_tgt_conn) in enumerate(n_src_n_tgt):
                np_tup[i, :n_src] = n_src_conn
                np_tup[i, n_src:] = n_tgt_conn
            tuples_efficient.append((exist, np_tup))
        self._write_to_cache(cache_path, tuples_efficient)

    def _iter_n_sources_targets(self, existence: NodeExistence = None):
        for exist in (self.iter_existence() if existence is None else [existence]):
            for n_src_conn in self.iter_sources(existence=exist):
                for n_tgt_conn in self.iter_targets(n_source=sum(n_src_conn), existence=exist):
                    yield n_src_conn, n_tgt_conn, exist

    def iter_existence(self, include_none=False):
        yield from self.existence_patterns.patterns
        if include_none:
            yield NodeExistence()

    def iter_sources(self, existence: NodeExistence = None):
        n_conn_override = existence.src_n_conn_override if existence is not None else None
        max_conn = self.get_max_src_appear(existence) if existence is not None else None
        yield from self._iter_conn_slots(self.src, n_conn_override=n_conn_override, max_conn=max_conn)

    def iter_targets(self, n_source, existence: NodeExistence = None):
        n_conn_override = existence.tgt_n_conn_override if existence is not None else None
        max_conn = self.get_max_tgt_appear(existence) if existence is not None else None
        yield from self._iter_conn_slots(
            self.tgt, is_src=False, n=n_source, n_conn_override=n_conn_override, max_conn=max_conn)

    def _iter_conn_slots(self, nodes: List[Node], is_src=True, n=None, n_conn_override: Dict[int, List[int]] = None,
                         max_conn: np.ndarray = None):
        """Iterate over all combinations of number of connections per nodes."""

        if max_conn is None:
            max_conn = self.get_max_src_appear() if is_src else self.get_max_tgt_appear()

        if n_conn_override is None:
            n_conn_override = {}

        # Get all possible number of connections for each node
        n_conns = []
        for i, node in enumerate(nodes):
            if i in n_conn_override:
                n_conns.append(n_conn_override[i])
            else:
                n_conns.append(self.get_node_conns(node, max_conn[i]))

        n_conns = tuple(tuple(n_conn) for n_conn in n_conns)
        return self._iter_conn_slots_inner_(n_conns, n)

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def _iter_conn_slots_inner_(n_conns, n):
        # If we have an amount constraint, use the matrix generation algorithm
        if n is not None:
            # Check if we have no connection slots available
            if len(n_conns) == 0:
                if n == 0:  # If zero connections are allowed, return an empty solution
                    return [tuple()]
                return []

            n_conn_max = [max(n_conn) if len(n_conn) > 0 else 0 for n_conn in n_conns]
            n_tgt_combs = count_src_to_target(n_src=n, n_src_to_target=tuple(n_conn_max))

            # Filter out invalid amounts
            for i_tgt, n_conn in enumerate(n_conns):
                tgt_n_conns = n_tgt_combs[:, i_tgt]
                n_invalid = set(tgt_n_conns) - set(n_conn)
                if len(n_invalid) > 0:
                    invalid_mask = np.zeros((len(tgt_n_conns),), dtype=bool)
                    for n in n_invalid:
                        invalid_mask |= tgt_n_conns == n
                    n_tgt_combs = n_tgt_combs[~invalid_mask, :]

            conn_slots = []
            for i in reversed(range(n_tgt_combs.shape[0])):  # Reversed for compatibility
                conn_slots.append(tuple(n_tgt_combs[i, :]))
            return conn_slots

        # Iterate over all combinations of number of connections
        conn_slots = []
        for n_conn_nodes in itertools.product(*n_conns):

            n_tot = sum(n_conn_nodes)
            if n is not None and n_tot != n:
                continue

            conn_slots.append(tuple(n_conn_nodes))
        return conn_slots

    def _get_node_settings(self):  # For numba-jit-compiled _check_conns
        def _get_node_settings(nodes: List[Node]):
            n_conns_check = [max(node.conns) for node in nodes if not node.max_inf]
            n_conns_check = 0 if len(n_conns_check) == 0 else max(n_conns_check)
            node_settings = np.zeros((len(nodes), 2+n_conns_check+1), dtype=np.int32)
            for i, node in enumerate(nodes):
                if node.max_inf:
                    node_settings[i, 0] = 1
                    node_settings[i, 1] = node.min_conns
                else:
                    node_settings[i, 2+np.array(node.conns)] = 1
            return node_settings

        if self._node_settings is None:
            self._node_settings = (_get_node_settings(self.src), _get_node_settings(self.tgt))
        return self._node_settings

    def _get_n_conn_override(self, existence: NodeExistence = None):
        if existence is None:
            existence = NodeExistence()

        def _make_n_conn_override(nodes: List[Node], override_map, node_settings):
            if override_map is None or len(override_map) == 0:
                return np.zeros((len(nodes), 0), dtype=np.int8)

            n_override_max = max(max([max(n_conns) for n_conns in override_map.values()]), node_settings.shape[1]-2)
            override_settings = np.zeros((len(nodes), n_override_max+1), dtype=np.int8)
            for i in range(len(nodes)):
                if i in override_map:
                    override_settings[i, np.array(override_map[i])] = 1
                else:
                    override_settings[i, :] = -1
            return override_settings

        if existence not in self._n_conn_override_map:
            src_settings, tgt_settings = self._get_node_settings()
            self._n_conn_override_map[existence] = (_make_n_conn_override(self.src, existence.src_n_conn_override, src_settings),
                                                    _make_n_conn_override(self.tgt, existence.tgt_n_conn_override, tgt_settings))
        return self._n_conn_override_map[existence]

    def validate_matrix(self, matrix: np.ndarray, existence: NodeExistence = None) -> bool:
        """Checks whether a connection matrix is valid"""

        if existence is None:
            existence = NodeExistence()

        src_node_settings, tgt_node_settings = self._get_node_settings()
        src_n_override, tgt_n_override = self._get_n_conn_override(existence=existence)
        max_src = self.get_max_src_appear(existence=existence)
        max_tgt = self.get_max_tgt_appear(existence=existence)
        max_conn_mat = self.get_max_conn_mat(existence)

        return _validate_matrix(matrix, max_conn_mat, src_node_settings,
                                tgt_node_settings, src_n_override, tgt_n_override, max_src, max_tgt)

    @staticmethod
    def _check_conns(n_conns: int, node: Node, max_conn: int) -> bool:
        if n_conns > max_conn:
            return False
        if node.max_inf:
            return node.min_conns <= n_conns
        return n_conns in node.conns

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def get_node_conns(node: Node, max_conn: int):
        if node.max_inf:
            slot_max_conn = max_conn
            return list(range(node.min_conns, slot_max_conn+1))

        return [c for c in node.conns if c <= max_conn]

    def count_all_matrices(self, max_by_existence=True) -> int:
        agg_matrix = self._load_agg_matrix_from_cache()
        if agg_matrix is not None:
            count_by_existence = {existence: matrix.shape[0] for existence, matrix in agg_matrix.items()}
        else:
            count_by_existence = {}
            for n_src_conn, n_tgt_conn, existence in self.iter_n_sources_targets():
                if existence not in count_by_existence:
                    count_by_existence[existence] = 0
                count_by_existence[existence] += self.count_matrices(n_src_conn, n_tgt_conn, existence)

        if max_by_existence:
            return max(count_by_existence.values())
        return sum(count_by_existence.values())

    def get_matrices_by_n_conn(self, n_src_conn, n_tgt_conn, existence: NodeExistence = None):
        return self._iter_matrices(n_src_conn, n_tgt_conn, existence=existence)

    def _iter_matrices(self, n_src_conn, n_tgt_conn, existence: NodeExistence = None):
        """
        Iterate over all permutations of the connections between the source and target objects, taking repetition and
        exclusion constraints into account.

        Inspired by [Selva2017: Patterns in System Architecture Decisions], we are using the assigning pattern and
        representing assignments using a matrix of n_src x n_tgt. Each index (i_src, j_tgt) represents a connection
        between a specific source and target, and the value in the matrix represents the number of connections.

        To create the different matrices, we have to adhere to three constraints:
        1. The sum of target connections (sum over columns) should be equal to the nr of times they occur in tgt_objs
        2. Ensure that in certain rows (sources) or columns (targets), all values are max 1 if no repetition is allowed
        3. Block certain connections from being made at all if we have excluded edges
        """
        if existence is None:
            existence = NodeExistence()

        # Check if we have any source connections
        if len(n_src_conn) == 0 or len(n_tgt_conn) == 0:
            return np.zeros((1, len(n_src_conn), len(n_tgt_conn)), dtype=int)

        # Generate matrices
        return self._get_matrices_numpy(n_src_conn, n_tgt_conn, existence)
        # return self._get_matrices_numba(n_src_conn, n_tgt_conn, existence)

    def count_matrices(self, n_src_conn, n_tgt_conn, existence: NodeExistence = None) -> int:
        """Count the number of permutations as would be generated by _iter_matrices."""
        if existence is None:
            existence = NodeExistence()

        # Count special cases
        n_mat = self._count_matrices_special(n_src_conn, n_tgt_conn, existence)
        if n_mat is not None:
            return n_mat

        # Count symmetric special cases
        n_mat = self._count_matrices_special(n_tgt_conn, n_src_conn, existence, is_switched=True)
        if n_mat is not None:
            return n_mat

        return self._get_matrices_numpy(n_src_conn, n_tgt_conn, existence, create_matrices=False)
        # return len(self._get_matrices_numba(n_src_conn, n_tgt_conn, existence, output_matrix=False))

    def _get_matrices_numba(self, n_src_conn, n_tgt_conn, existence: NodeExistence, output_matrix=True):
        n_src_conn = np.array(n_src_conn, dtype=np.int64)
        n_tgt_conn = np.array(n_tgt_conn, dtype=np.int64)
        max_conn_mat = self.get_max_conn_mat(existence)
        matrices = yield_matrices(
            len(n_src_conn), len(n_tgt_conn), np.array(n_src_conn, dtype=np.int64),
            np.array(n_tgt_conn, dtype=np.int64), max_conn_mat)

        if output_matrix:
            if len(matrices) == 0:
                return np.empty((0, len(n_src_conn), len(n_tgt_conn)), dtype=int)
            elif len(matrices) == 1:
                return np.array([matrices[0]])
            return np.array(matrices)
        return matrices

    def _count_matrices_special(self, n_src_conn, n_tgt_conn, existence: NodeExistence, is_switched=False) \
            -> Optional[int]:
        # No connections
        n_src_total = sum(n_src_conn)
        n_tgt_total = sum(n_tgt_conn)
        if n_src_total == 0 or n_tgt_total == 0:
            return 1

        # 1 outgoing connection
        if len(n_src_conn) == 1 and n_src_conn[0] == 1:
            n_mat = 0
            for n_tgt in n_tgt_conn:
                if n_tgt > 0:
                    n_mat += 1
            return n_mat

        # Get max conn
        max_conn = self.get_max_conn_mat(existence)
        if is_switched:
            max_conn = max_conn.T

        # One of the src nodes has same nr of target connections
        for i_src, n_src in enumerate(n_src_conn):
            if n_src == n_tgt_total:
                # Check if any of the connection nr constraints are violated
                if np.any(n_tgt_conn > max_conn[i_src, :]):
                    return 0
                return 1

    def _get_matrices_numpy(self, n_src_conn, n_tgt_conn, existence: NodeExistence, create_matrices=True):
        max_conn_mat = self.get_max_conn_mat(existence)
        return _get_matrices(
            tuple(n_src_conn), tuple(n_tgt_conn), tuple(max_conn_mat.ravel()), create_matrices=create_matrices)

    def _get_max_conn(self) -> int:
        """
        Calculate the maximum number of connections that will be formed, taking infinite connections into account.
        """

        def _max(nodes):
            if len(nodes) == 0:
                return 0
            return sum([self.get_node_max_conn(node) for node in nodes])

        return max([_max(self.src), _max(self.tgt)])

    @staticmethod
    def get_node_max_conn(node: Node, n_min_max_inf: int = 1):
        return max(n_min_max_inf, node.min_conns) if node.max_inf else max(node.conns)

    @staticmethod
    def _get_min_conns(node: Node) -> int:
        if node.max_inf:
            return node.min_conns
        return node.conns[0]


@numba.njit()
def _validate_matrix(matrix: np.ndarray, max_conn_mat: np.ndarray, src_node_settings: np.ndarray,
                     tgt_node_settings: np.ndarray, src_n_override: np.ndarray, tgt_n_override: np.ndarray,
                     max_src: np.ndarray, max_tgt: np.ndarray) -> bool:
    # Check repeated or blocked connections
    if np.any(matrix > max_conn_mat):
        return False

    # Check source connections
    for i in range(matrix.shape[0]):
        n_src = np.sum(matrix[i, :])
        if n_src < src_n_override.shape[1]:
            override_ok = src_n_override[i, n_src]
            if override_ok == 0:
                return False
            elif override_ok == 1:
                continue
        if not _check_conns(n_src, src_node_settings[i, :], max_src[i]):
            return False

    # Check target connections
    for i in range(matrix.shape[1]):
        n_tgt = np.sum(matrix[:, i])
        if n_tgt < tgt_n_override.shape[1]:
            override_ok = tgt_n_override[i, n_tgt]
            if override_ok == 0:
                return False
            elif override_ok == 1:
                continue
        if not _check_conns(n_tgt, tgt_node_settings[i, :], max_tgt[i]):
            return False

    return True


@numba.njit()
def _check_conns(n_conns, node_settings: np.ndarray, max_conn: int) -> bool:  # See get_node_settings for data structure
    if n_conns > max_conn:
        return False
    if node_settings[0]:  # 0 = max_inf
        return node_settings[1] <= n_conns  # 1 = min_conns

    if n_conns+2 >= len(node_settings):
        return False
    return node_settings[n_conns+2] == 1


@numba.njit()
def yield_matrices(n_src: int, n_tgt: int, n_src_conn, n_tgt_conn, max_conn):
    init_matrix = np.zeros((n_src, n_tgt), dtype=np.int64)
    init_tgt_sum = np.zeros((n_tgt,), dtype=np.int64)
    last_src_idx = n_src-1
    last_tgt_idx = n_tgt-1

    return _branch_matrices(0, init_matrix, init_tgt_sum, n_src_conn, n_tgt_conn, max_conn, n_tgt,
                            last_src_idx, last_tgt_idx)


@numba.njit(parallel=True)
def _branch_matrices(i_src, current_matrix, current_sum, n_src_conn, n_tgt_conn, max_conn, n_tgt,
                     last_src_idx, last_tgt_idx):
    # Get total connections from this source
    n_total_conns = n_src_conn[i_src]

    # Determine the amount of target connections available (per target node) for the current source node
    n_tgt_conn_avbl = n_tgt_conn-current_sum
    for i_tgt in range(len(n_tgt_conn_avbl)):
        if n_tgt_conn_avbl[i_tgt] > max_conn[i_src, i_tgt]:
            n_tgt_conn_avbl[i_tgt] = max_conn[i_src, i_tgt]

    # Get number of minimum connections per target in order to be able to distribute all outgoing connections
    # starting from the first target
    reverse_cum_max_conn = np.cumsum(n_tgt_conn_avbl[::-1])[::-1]
    n_tgt_conn_min = (n_total_conns-reverse_cum_max_conn)[1:]

    # Loop over different source-to-targets connection patterns
    branched_matrices = []
    init_tgt_conns = np.zeros((n_tgt,), dtype=np.int64)
    for tgt_conns in _get_branch_tgt_conns(
            init_tgt_conns, 0, 0, n_total_conns, n_tgt_conn_avbl, n_tgt_conn_min, last_tgt_idx):
        # Modify matrix
        next_matrix = current_matrix.copy()
        next_matrix[i_src, :] = tgt_conns

        # Check if we can stop branching out
        if i_src == last_src_idx:
            branched_matrices.append(next_matrix)
            continue

        # Modify matrix sum
        next_sum = current_sum+tgt_conns

        # Branch out for next src node
        for branched_matrix in _branch_matrices(i_src+1, next_matrix, next_sum, n_src_conn, n_tgt_conn, max_conn,
                                                n_tgt, last_src_idx, last_tgt_idx):
            branched_matrices.append(branched_matrix)

    return branched_matrices


@numba.njit()
def _get_branch_tgt_conns(tgt_conns, i_tgt, n_conn_set, n_total_conns, n_tgt_conn_avbl, n_tgt_conn_min, last_tgt_idx):
    # Check if we already have distributed all connections
    if n_conn_set == n_total_conns:
        return [tgt_conns]
    sub_tgt_conns = []

    # Branch: add another connection to the current target index, if below max for this target
    if tgt_conns[i_tgt] < n_tgt_conn_avbl[i_tgt]:
        next_tgt_conns = tgt_conns.copy()
        next_tgt_conns[i_tgt] += 1
        for sub_tgt_conn in _get_branch_tgt_conns(
                next_tgt_conns, i_tgt, n_conn_set+1, n_total_conns, n_tgt_conn_avbl, n_tgt_conn_min, last_tgt_idx):
            sub_tgt_conns.append(sub_tgt_conn)

    # Branch: move to next target, if we are not at the end and if we are above the minimum connections
    if i_tgt < last_tgt_idx and n_conn_set >= n_tgt_conn_min[i_tgt]:
        for sub_tgt_conn in _get_branch_tgt_conns(
                tgt_conns, i_tgt+1, n_conn_set, n_total_conns, n_tgt_conn_avbl, n_tgt_conn_min, last_tgt_idx):
            sub_tgt_conns.append(sub_tgt_conn)

    return sub_tgt_conns


@functools.lru_cache(maxsize=None)
def _get_matrices(n_src_conn, n_tgt_conn, max_conn_mat_tup, create_matrices=True):
    """Recursive column-wise (target-wise)"""
    n_src, n_tgt = len(n_src_conn), len(n_tgt_conn)
    max_conn_mat = np.array(max_conn_mat_tup).reshape(n_src, n_tgt)

    # Determine number of src connections to first target
    n_target_from_src = np.minimum(n_src_conn, max_conn_mat[:, 0])

    # Get combinations of sources to target assignments
    create_here = create_matrices or len(n_tgt_conn) > 1
    n_taken_combs = count_src_to_target(
        n_tgt_conn[0], tuple(n_target_from_src), create_matrices=create_here)
    if (n_taken_combs.shape[0] if create_here else n_taken_combs) == 0:
        return np.empty((0, n_src, n_tgt), dtype=int) if create_matrices else 0

    # Only one target left to connect to
    if len(n_tgt_conn) == 1:
        if create_matrices:
            return n_taken_combs.reshape((n_taken_combs.shape[0], n_src, 1))
        return n_taken_combs

    # Get combinations for first column
    matrices = []
    count = 0
    n_src_conn_arr = np.array(n_src_conn)
    for i_comb, n_from_src in enumerate(n_taken_combs):
        n_src_conn_remain = n_src_conn_arr-n_from_src
        next_mc_mat = tuple(max_conn_mat[:, 1:].ravel())

        # Get combinations for further columns
        next_matrices = _get_matrices(
            tuple(n_src_conn_remain), n_tgt_conn[1:], next_mc_mat, create_matrices=create_matrices)

        if not create_matrices:
            count += next_matrices
            continue

        matrices_i = np.empty((next_matrices.shape[0], n_src, n_tgt), dtype=int)
        matrices_i[:, :, 0] = n_from_src
        matrices_i[:, :, 1:] = next_matrices
        matrices.append(matrices_i)

    if not create_matrices:
        return count
    return matrices[0] if len(matrices) == 1 else np.row_stack(matrices)


@functools.lru_cache(maxsize=None)
# @numba.njit()
def count_src_to_target(n_src, n_src_to_target, create_matrices=True):
    # Prepare slot duplication definition
    n_flat_slots = np.sum(n_src_to_target)
    slot_i_tgt = np.empty((n_flat_slots,), dtype=np.int64)
    n_dup_next = np.zeros((n_flat_slots,), dtype=np.int64)
    i_slot = 0
    for i_tgt, n in enumerate(n_src_to_target):
        slot_i_tgt[i_slot:i_slot+n] = i_tgt
        if n > 1:
            n_dup_next[i_slot] = n-1
        i_slot += n
    n_dup_next = tuple(n_dup_next)  # For cached version

    # Count combinations for this source node
    count, combinations = count_n_pool_take(len(n_dup_next), n_src, n_dup_next)
    if not create_matrices:
        return count

    # Find nr of connections to slots
    n_taken_combs = np.zeros((count, len(n_src_to_target)), dtype=np.int64)
    for i_tgt, n in enumerate(n_src_to_target):
        if n == 0:
            continue

        conn_comb_tgt = combinations[:, slot_i_tgt == i_tgt]
        n_taken_combs[:, i_tgt] = conn_comb_tgt[:, 0] if n == 1 else np.sum(conn_comb_tgt, axis=1)

    return n_taken_combs


@functools.lru_cache(maxsize=None)
# @numba.njit()
def count_n_pool_take(n_pool, n_take, n_dup_next):
    # Special cases: at the edges of n_take wrt n_pool, we have only 1 combination (duplications don't matter)
    # Example: take 0 from 4 --> 0000; take 4 from 4 --> 1111
    if n_take == 0 or n_take == n_pool:
        combinations = np.zeros((1, n_pool), dtype=np.int64) if n_take == 0 else np.ones((1, n_pool), dtype=np.int64)
        return 1, combinations
    elif n_take < 0 or n_take > n_pool:
        return 0, np.zeros((0, n_pool), dtype=np.int64)

    # We are symmetric wrt n_take, and since the algorithm is a bit faster for lower numbers of n_take, we swap the
    # value if we are higher than half of n_pool
    is_swapped = False
    if n_pool > 1 and n_take > n_pool*.5:
        n_take = n_pool-n_take
        is_swapped = True

    # Get initial combinations matrix
    combinations = _combinations_matrix(n_pool, n_take)

    # Process duplication constraints
    if np.sum(n_dup_next) > 0:
        mask_invalid_dup = np.zeros((combinations.shape[0],), dtype=bool)  # For cached
        # mask_invalid_dup = np.zeros((combinations.shape[0],), dtype=numba.types.bool_)  # For numba
        for i, n_dup in enumerate(n_dup_next):
            if n_dup == 0:
                continue

            # Valid rows have uniformly decreasing or equal number of connections (e.g. 1,1,0, but not 1,0,1 or 0,1,1)
            n_dup_diff = np.diff(combinations[:, i:i+n_dup+1], axis=1)  # For cached
            # n_dup_diff = combinations[:, i+1:i+n_dup+1]-combinations[:, i:i+n_dup]  # For numba
            for j in range(n_dup_diff.shape[1]):
                mask_invalid_dup |= n_dup_diff[:, j] > 0

        # Remove invalid rows
        combinations = combinations[~mask_invalid_dup, :]

    if is_swapped:
        # The consequence of this is that the duplicated slots are filled from the end instead of from the front, but
        # for adding/counting that shouldn't matter
        combinations = (1-combinations)[::-1, :]

    return combinations.shape[0], combinations


@functools.lru_cache(maxsize=None)
# @numba.njit()
def _combinations_matrix(p, r):
    # Based on nump2 of https://stackoverflow.com/a/42202157
    # Prepare combinations matrix and first column
    n0 = p-r+1
    comb_mat = np.zeros((n0, p), dtype=np.int64)
    comb_mat[:n0, :n0] = np.eye(n0)
    n_rep = np.arange(p-r+1)
    # comb_mat[n_rep, n_rep] = 1

    for j in range(1, r):
        # Determine number of repetitions
        reps = (p-r+j) - n_rep
        comb_mat = np.repeat(comb_mat, reps, axis=0)  # For cached
        # comb_mat_rep = np.empty((np.sum(reps), comb_mat.shape[1]), dtype=np.int64)  # For numba
        # i_in_mat = 0
        # for i_rep, n_rep in enumerate(reps):
        #     comb_mat_rep[i_in_mat:i_in_mat+n_rep, :] = comb_mat[i_rep]
        #     i_in_mat += n_rep
        # comb_mat = comb_mat_rep

        ind = np.add.accumulate(reps)  # For cached
        # ind = _add_accumulate(reps)  # For numba

        # Get columns to set to True
        n_rep = np.ones((comb_mat.shape[0],), dtype=np.int64)
        n_rep[ind[:-1]] = 1-reps[1:]
        n_rep[0] = j
        n_rep[:] = np.add.accumulate(n_rep[:])  # For cached
        # n_rep[:] = _add_accumulate(n_rep[:])  # For numba

        comb_mat[np.arange(comb_mat.shape[0]), n_rep] = 1  # For cached
        # for i_rep, i_col in enumerate(n_rep):  # For numba
        #     comb_mat[i_rep, i_col] = True

    return comb_mat


@numba.njit()
def _add_accumulate(arr):
    arr_acc = arr.copy()
    for i in range(1, len(arr)):
        arr_acc[i:] = arr_acc[i-1]+arr[i]
    return arr_acc



import itertools
import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.lazy_encoding import *
from adsg_core.optimization.assign_enc.lazy.imputation.delta import *

__all__ = ['LazyClosestImputer']


class LazyClosestImputer(LazyDeltaImputer):
    """Imputes by looking for the closest valid design vector based on euclidean or Mahattan distance.
    Eagerly evaluates itertools.product, so might need too much memory/time."""

    def __init__(self, euclidean=True, n_max_tries=None):
        self.euclidean = euclidean
        super().__init__(n_max_tries=n_max_tries)

    def yield_dv_delta_product(self, delta_values):
        # Sort all delta values by distance (skip first as it has a delta of 0)
        dist_func = self._calc_dist_euclidean if self.euclidean else self._calc_dist_manhattan
        dv_deltas = [np.array(dv_delta) for dv_delta in itertools.product(*delta_values)][1:]  # Skip first as it has a delta of 0
        dv_dist = dist_func(np.array(dv_deltas))

        # Sort by distance
        i_dv_dist = np.argsort(dv_dist)
        for i_dv in i_dv_dist:
            yield dv_deltas[i_dv]

    @staticmethod
    def _calc_dist_euclidean(deltas: np.ndarray) -> np.ndarray:
        return np.sqrt(np.sum(deltas**2, axis=1))

    @staticmethod
    def _calc_dist_manhattan(deltas: np.ndarray) -> np.ndarray:
        return np.sum(np.abs(deltas), axis=1)

    def __repr__(self):
        return f'{self.__class__.__name__}(euclidean={self.euclidean})'

    def __str__(self):
        return f'Closest {"Euc" if self.euclidean else "Mht"} Imp'



import numpy as np
from typing import *
from adsg_core.optimization.assign_enc.lazy_encoding import *

__all__ = ['LazyConstraintViolationImputer']


class LazyConstraintViolationImputer(LazyImputer):
    """Does not do imputation, but sends a matrix with -1 as values to indicate an invalid design point.
    The optimization problem should represent this as a violated constraint."""

    def impute(self, vector: DesignVector, matrix: Optional[np.ndarray], existence: NodeExistence,
               tried_vectors: set = None) -> Tuple[DesignVector, np.ndarray]:
        invalid_matrix = -np.ones((len(self._matrix_gen.src), len(self._matrix_gen.tgt)), dtype=int)
        return vector, invalid_matrix

    def _impute(self, *args, **kwargs):
        raise RuntimeError

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def __str__(self):
        return f'Constr Vio Imp'
